'''
mesoSPIM Waveform Generator - Creates and allows control of waveform generation.
'''
import os
import numpy as np
import csv
import time

import logging
logger = logging.getLogger(__name__)

'''National Instruments Imports'''
import nidaqmx
from nidaqmx.constants import AcquisitionType, TaskMode
from nidaqmx.constants import LineGrouping, DigitalWidthUnits
from nidaqmx.types import CtrTime

'''mesoSPIM imports'''
from .mesoSPIM_State import mesoSPIM_StateSingleton
from .utils.waveforms import single_pulse, tunable_lens_ramp, sawtooth, square

from PyQt5 import QtCore

class mesoSPIM_WaveFormGenerator(QtCore.QObject):
    '''This class contains the microscope state

    Any access to this global state should only be done via signals sent by
    the responsible class for actually causing that state change in hardware.

    '''
    sig_update_gui_from_state = QtCore.pyqtSignal(bool) # -> mesoSPIM_Core.sig_update_gui_from_state -> MainWindow.enable_gui_updates_from_state

    def __init__(self, parent):
        super().__init__()
        self.cfg = parent.cfg
        self.parent = parent # mesoSPIM_Core object
        self.state = mesoSPIM_StateSingleton()
        self.parent.sig_save_etl_config.connect(self.save_etl_parameters_to_csv)
        cfg_file = self.parent.read_config_parameter('ETL_cfg_file', self.cfg.startup)
        self.state['ETL_cfg_file'] = cfg_file
        self.update_etl_parameters_from_csv(cfg_file, self.state['laser'], self.state['zoom'])
        #logger.debug('Thread ID at Startup: '+str(int(QtCore.QThread.currentThreadId())))
        self.state['galvo_l_amplitude'] = self.parent.read_config_parameter('galvo_l_amplitude', self.cfg.startup)
        self.state['galvo_r_amplitude'] = self.parent.read_config_parameter('galvo_r_amplitude', self.cfg.startup)
        self.state['galvo_l_frequency'] = self.parent.read_config_parameter('galvo_l_frequency', self.cfg.startup)
        self.state['galvo_r_frequency'] = self.parent.read_config_parameter('galvo_r_frequency', self.cfg.startup)
        self.state['galvo_l_offset'] = self.parent.read_config_parameter('galvo_l_offset', self.cfg.startup)
        self.state['galvo_r_offset'] = self.parent.read_config_parameter('galvo_r_offset', self.cfg.startup)
        self.state['max_laser_voltage'] = self.parent.read_config_parameter('max_laser_voltage', self.cfg.startup)
        self.config_check()

    def config_check(self):
        '''Check config file for old/wrong/deprecated pieces'''
        if hasattr(self.cfg, 'laser_designation'):
            print("INFO: Config file: The 'laser_designation' dictionary is obsolete, you can remove it.")
        if hasattr(self.cfg, 'galvo_etl_designation'):
            print("INFO: Config file: The 'galvo_etl_designation' dictionary is obsolete, you can remove it.")
        laser_task_line_start = int(self.cfg.acquisition_hardware['laser_task_line'].split(':')[0][-1])
        laser_task_line_end = int(self.cfg.acquisition_hardware['laser_task_line'].split(':')[1])
        if (laser_task_line_end - laser_task_line_start + 1) != len(self.cfg.laserdict):
            raise ValueError(f"Config file: number of AO lines in 'laser_task_line' "
                             f"({self.cfg.acquisition_hardware['laser_task_line']}) "
                             f"must be equal to num(lasers) in 'laserdict' ({len(self.cfg.laserdict)}). "
                             f"Check assignment of AO channels in 'laser_task_line'.")

    @QtCore.pyqtSlot(dict)
    def state_request_handler(self, dict):
        for key, value in zip(dict.keys(), dict.values()):
            if key in ('samplerate',
                       'sweeptime',
                       'intensity',
                       'etl_l_delay_%',
                       'etl_l_ramp_rising_%',
                       'etl_l_ramp_falling_%',
                       'etl_l_amplitude',
                       'etl_l_offset',
                       'etl_r_delay_%',
                       'etl_r_ramp_rising_%',
                       'etl_r_ramp_falling_%',
                       'etl_r_amplitude',
                       'etl_r_offset',
                       'galvo_l_frequency',
                       'galvo_l_amplitude',
                       'galvo_l_offset',
                       'galvo_l_duty_cycle',
                       'galvo_l_phase',
                       'galvo_r_frequency',
                       'galvo_r_amplitude',
                       'galvo_r_offset',
                       'galvo_r_duty_cycle',
                       'galvo_r_phase',
                       'laser_l_delay_%',
                       'laser_l_pulse_%',
                       'laser_l_max_amplitude',
                       'laser_r_delay_%',
                       'laser_r_pulse_%',
                       'laser_r_max_amplitude',
                        'laser',
                       'camera_delay_%',
                       'camera_pulse_%',
                       'shutterconfig',
                       'zoom',
                       ):
                ''' Notify GUI about the change '''
                self.sig_update_gui_from_state.emit(True)
                self.state[key] = value
                self.sig_update_gui_from_state.emit(False)
                self.create_waveforms()
            elif key == 'ETL_cfg_file':
                self.state[key] = value
                self.update_etl_parameters_from_csv(value, self.state['laser'], self.state['zoom'])
            elif key == 'set_etls_according_to_zoom':
                self.update_etl_parameters_from_zoom(value)
            elif key == 'set_etls_according_to_laser':
                self.update_etl_parameters_from_laser(value)

            elif key == 'state':
                if value == 'live':
                    logger.debug('Thread ID during live: '+str(int(QtCore.QThread.currentThreadId())))

    def calculate_samples(self):
        samplerate, sweeptime = self.state.get_parameter_list(['samplerate', 'sweeptime'])
        self.samples = int(samplerate*sweeptime)

    def create_waveforms(self):
        logger.info("waveforms updated")
        self.calculate_samples()
        self.create_etl_waveforms()
        self.create_galvo_waveforms()
        '''Bundle everything'''
        self.bundle_galvo_and_etl_waveforms()
        self.create_laser_waveforms()

    def create_etl_waveforms(self):
        samplerate, sweeptime = self.state.get_parameter_list(['samplerate', 'sweeptime'])
        etl_l_delay, etl_l_ramp_rising, etl_l_ramp_falling, etl_l_amplitude, etl_l_offset = \
            self.state.get_parameter_list(['etl_l_delay_%', 'etl_l_ramp_rising_%', 'etl_l_ramp_falling_%', 'etl_l_amplitude','etl_l_offset'])
        etl_r_delay, etl_r_ramp_rising, etl_r_ramp_falling, etl_r_amplitude, etl_r_offset = \
            self.state.get_parameter_list(['etl_r_delay_%', 'etl_r_ramp_rising_%', 'etl_r_ramp_falling_%', 'etl_r_amplitude', 'etl_r_offset'])

        self.etl_l_waveform = tunable_lens_ramp(samplerate = samplerate,
                                                sweeptime = sweeptime,
                                                delay = etl_l_delay,
                                                rise = etl_l_ramp_rising,
                                                fall = etl_l_ramp_falling,
                                                amplitude = etl_l_amplitude,
                                                offset = etl_l_offset)

        self.etl_r_waveform = tunable_lens_ramp(samplerate = samplerate,
                                                sweeptime = sweeptime,
                                                delay = etl_r_delay,
                                                rise = etl_r_ramp_rising,
                                                fall = etl_r_ramp_falling,
                                                amplitude = etl_r_amplitude,
                                                offset = etl_r_offset)
        # freeze AO channel which is not in use, to reduce heating and increase ETL lifetime
        if self.state['shutterconfig'] == 'Left':
            self.etl_r_waveform[:] = etl_r_offset
            logger.debug("Right arm frozen")
        elif self.state['shutterconfig'] == 'Right':
            self.etl_l_waveform[:] = etl_l_offset
            logger.debug("Left arm frozen")
        else:
            pass

    def create_galvo_waveforms(self):
        samplerate, sweeptime = self.state.get_parameter_list(['samplerate','sweeptime'])

        galvo_l_frequency, galvo_l_amplitude, galvo_l_offset, galvo_l_duty_cycle, galvo_l_phase =\
        self.state.get_parameter_list(['galvo_l_frequency', 'galvo_l_amplitude', 'galvo_l_offset',
        'galvo_l_duty_cycle', 'galvo_l_phase'])

        galvo_r_frequency, galvo_r_amplitude, galvo_r_offset, galvo_r_duty_cycle, galvo_r_phase =\
        self.state.get_parameter_list(['galvo_r_frequency', 'galvo_r_amplitude', 'galvo_r_offset',
        'galvo_r_duty_cycle', 'galvo_r_phase'])

        '''Create Galvo waveforms:'''
        self.galvo_l_waveform = sawtooth(samplerate = samplerate,
                                         sweeptime = sweeptime,
                                         frequency = galvo_l_frequency,
                                         amplitude = galvo_l_amplitude,
                                         offset = galvo_l_offset,
                                         dutycycle = galvo_l_duty_cycle,
                                         phase = galvo_l_phase)

        self.galvo_r_waveform = sawtooth(samplerate = samplerate,
                                         sweeptime = sweeptime,
                                         frequency = galvo_r_frequency,
                                         amplitude = galvo_r_amplitude,
                                         offset = galvo_r_offset,
                                         dutycycle = galvo_r_duty_cycle,
                                         phase = galvo_r_phase)

        # freeze AO channel which is not in use, to reduce heating and increase galvo lifetime
        if self.state['shutterconfig'] == 'Left':
            self.galvo_r_waveform[:] = galvo_r_offset
        elif self.state['shutterconfig'] == 'Right':
            self.galvo_l_waveform[:] = galvo_l_offset
        else:
            pass

    def create_laser_waveforms(self):
        samplerate, sweeptime = self.state.get_parameter_list(['samplerate','sweeptime'])

        laser_l_delay, laser_l_pulse, max_laser_voltage, intensity = \
        self.state.get_parameter_list(['laser_l_delay_%','laser_l_pulse_%', 'max_laser_voltage', 'intensity'])

        '''Create zero waveforms for the lasers'''
        self.zero_waveform = np.zeros((self.samples))

        '''Update the laser intensity waveform'''
        '''This could be improved: create a list with as many zero arrays as analog out lines for ETL and Lasers'''
        self.laser_waveform_list = [self.zero_waveform for i in self.cfg.laserdict]

        ''' Conversion from % to V of the intensity:'''
        laser_voltage = max_laser_voltage * intensity / 100

        self.laser_template_waveform = single_pulse(samplerate = samplerate,
                                                    sweeptime = sweeptime,
                                                    delay = laser_l_delay,
                                                    pulsewidth = laser_l_pulse,
                                                    amplitude = laser_voltage,
                                                    offset = 0)

        '''The key: replace the waveform in the waveform list with this new template'''
        assert sorted(list(self.cfg.laserdict.keys())) == list(self.cfg.laserdict.keys()), f"Error: laserdict keys in config file must be alphanumerically sorted: {self.cfg.laserdict.keys()}"
        current_laser_index = sorted(list(self.cfg.laserdict.keys())).index(self.state['laser'])
        self.laser_waveform_list[current_laser_index] = self.laser_template_waveform
        self.laser_waveforms = np.stack(self.laser_waveform_list)

    def bundle_galvo_and_etl_waveforms(self):
        """ Stacks the Galvo and ETL waveforms into a numpy array adequate for
        the NI cards.

        In here, the assignment of output channels of the Galvo / ETL card to the
        corresponding output channel is hardcoded: This could be improved.
        """
        self.galvo_and_etl_waveforms = np.stack((self.galvo_l_waveform,
                                                 self.galvo_r_waveform,
                                                 self.etl_l_waveform,
                                                 self.etl_r_waveform))

    def update_etl_parameters_from_zoom(self, zoom):
        """ Little helper method: Because the mesoSPIM core is not handling
        the serial Zoom connection. """
        laser = self.state['laser']
        etl_cfg_file = self.state['ETL_cfg_file']
        self.update_etl_parameters_from_csv(etl_cfg_file, laser, zoom)

    def update_etl_parameters_from_laser(self, laser):
        """ Little helper method: Because laser changes need an ETL parameter update """
        zoom = self.state['zoom']
        etl_cfg_file = os.path.join(self.parent.package_directory, self.state['ETL_cfg_file'])
        self.update_etl_parameters_from_csv(etl_cfg_file, laser, zoom)

    def update_etl_parameters_from_csv(self, cfg_path, laser, zoom):
        """ Updates the internal ETL left/right offsets and amplitudes from the
        values in the ETL csv files

        The .csv file needs to contain the follwing columns:

        Wavelength
        Zoom
        ETL-Left-Offset
        ETL-Left-Amp
        ETL-Right-Offset
        ETL-Right-Amp
        """
        self.sig_update_gui_from_state.emit(True)
        full_path = os.path.join(self.parent.package_directory, cfg_path)
        with open(full_path) as file:
            reader = csv.DictReader(file,delimiter=';')
            #print('opened csv')
            match_found = False
            for row in reader:
                if row['Wavelength'] == laser and row['Zoom'] == zoom:
                    match_found = True
                    # Some diagnostic tracing statements
                    # print(row)
                    ''' updating internal state '''
                    etl_l_offset = float(row['ETL-Left-Offset'])
                    etl_l_amplitude = float(row['ETL-Left-Amp'])
                    etl_r_offset = float(row['ETL-Right-Offset'])
                    etl_r_amplitude = float(row['ETL-Right-Amp'])

                    parameter_dict = {'etl_l_offset': etl_l_offset,
                                      'etl_l_amplitude' : etl_l_amplitude,
                                      'etl_r_offset' : etl_r_offset,
                                      'etl_r_amplitude' : etl_r_amplitude}

                    '''  Now the GUI needs to be updated '''
                    logger.info('Parameters set from csv')
                    self.state.set_parameters(parameter_dict)

        if match_found:
            '''Update waveforms with the new parameters'''
            self.create_waveforms()
            self.sig_update_gui_from_state.emit(False)
        else:
            logger.error(f"Laser {laser}, zoom {zoom} pair not found in file {cfg_path}. \n "
                         f"Check is file {cfg_path} is configured correctly for your laser and zoom dictionary")

    @QtCore.pyqtSlot()
    def save_etl_parameters_to_csv(self):
        """ Saves the current ETL left/right offsets and amplitudes from the
        values to the ETL csv files

        The .csv file needs to contain the following columns:

        Wavelength
        Zoom
        ETL-Left-Offset
        ETL-Left-Amp
        ETL-Right-Offset
        ETL-Right-Amp

        Creates a temporary cfg file with the ending _tmp
        """

        etl_cfg_file, laser, zoom, etl_l_offset, etl_l_amplitude, etl_r_offset, etl_r_amplitude = \
        self.state.get_parameter_list(['ETL_cfg_file', 'laser', 'zoom',
        'etl_l_offset', 'etl_l_amplitude', 'etl_r_offset','etl_r_amplitude'])

        '''Temporary filepath'''
        etl_cfg_file = os.path.join(self.parent.package_directory, etl_cfg_file)
        tmp_etl_cfg_file = etl_cfg_file+'_tmp'
        with open(etl_cfg_file,'r') as input_file, open(tmp_etl_cfg_file,'w') as outputfile:
            reader = csv.DictReader(input_file,delimiter=';')
            #print('created reader')
            fieldnames = ['Objective',
                          'Wavelength',
                          'Zoom',
                          'ETL-Left-Offset',
                          'ETL-Left-Amp',
                          'ETL-Right-Offset',
                          'ETL-Right-Amp']

            writer = csv.DictWriter(outputfile,fieldnames=fieldnames,dialect='excel',delimiter=';')
            writer.writeheader()
            for row in reader:
                if row['Wavelength'] == laser and row['Zoom'] == zoom:
                        writer.writerow({'Objective' : '1x',
                                         'Wavelength' : laser,
                                         'Zoom' : zoom,
                                         'ETL-Left-Offset' : etl_l_offset,
                                         'ETL-Left-Amp' : etl_l_amplitude,
                                         'ETL-Right-Offset' : etl_r_offset,
                                         'ETL-Right-Amp' : etl_r_amplitude,
                                         })
                else:
                    writer.writerow(row)
            writer.writerows(reader)
        os.remove(etl_cfg_file)
        os.rename(tmp_etl_cfg_file, etl_cfg_file)

    def create_tasks(self):
        """Creates a tasks for the mesoSPIM:

        These are:
        - the master trigger task, a digital out task that only provides a trigger pulse for the others
        - the camera trigger task, a counter task that triggers the camera in lightsheet mode
        - the stage trigger task, a counter task that provides a TTL trigger for stages that allow triggered movement (e.g. ASI stages)
        - the galvo and ETL task (analog out) that controls the left & right galvos for creation of
          the light-sheet and shadow avoidance
        - the aser task (analog out) that controls all the laser intensities (Laser should only
          be on when the camera is acquiring) and the left/right ETL waveforms.
          This task is bundled with galvo-ETL task if a single DAQmx card is used, because multifunction DAQmx devices
          can only run only 1 AO hardware-timed task at a time (https://knowledge.ni.com/KnowledgeArticleDetails?id=kA00Z0000019KWYSA2&l=en-CH)
        """
        ah = self.cfg.acquisition_hardware

        self.calculate_samples()
        samplerate, sweeptime = self.state.get_parameter_list(['samplerate','sweeptime'])
        samples = self.samples
        camera_pulse_percent, camera_delay_percent = self.state.get_parameter_list(['camera_pulse_%','camera_delay_%'])
        self.master_trigger_task = nidaqmx.Task()
        self.camera_trigger_task = nidaqmx.Task()
        if self.cfg.stage_parameters['stage_type'] in {'TigerASI'}:
            self.stage_trigger_task = nidaqmx.Task()

        # Check if 1 or 2 DAQ cards are used for AO waveform generation
        self.ao_cards = 1 if ah['galvo_etl_task_line'].split('/')[-2] == ah['laser_task_line'].split('/')[-2] else 2

        if self.ao_cards == 1:
            # These AO tasks than must be bundled into one task if a single DAQmx card is used (e.g. PXI-6733)
            self.galvo_etl_laser_task = nidaqmx.Task()
        else:
            self.galvo_etl_task = nidaqmx.Task()
            self.laser_task = nidaqmx.Task()

        '''Housekeeping: Setting up the DO master trigger task'''
        self.master_trigger_task.do_channels.add_do_chan(ah['master_trigger_out_line'],
                                                         line_grouping=LineGrouping.CHAN_FOR_ALL_LINES)

        '''Calculate camera high time and initial delay:
        Disadvantage: high time and delay can only be set after a task has been created
        '''
        self.camera_high_time = camera_pulse_percent*0.01*sweeptime
        self.camera_delay = camera_delay_percent*0.01*sweeptime

        '''Housekeeping: Setting up the counter task for the camera trigger'''
        self.camera_trigger_task.co_channels.add_co_pulse_chan_time(ah['camera_trigger_out_line'],
                                                                    high_time=self.camera_high_time,
                                                                    initial_delay=self.camera_delay)

        self.camera_trigger_task.triggers.start_trigger.cfg_dig_edge_start_trig(ah['camera_trigger_source'])

        '''Housekeeping: Setting up the counter task for the stage TTL trigger for certain stages'''
        if self.cfg.stage_parameters['stage_type'] in {'TigerASI'}:
            assert hasattr(self.cfg, 'asi_parameters'), "Config file with stage 'TigerASI' must contain 'asi_parameters' dictionary"
            trig_line = self.parent.read_config_parameter('stage_trigger_out_line', self.cfg.asi_parameters)
            trig_source = self.parent.read_config_parameter('stage_trigger_source', self.cfg.asi_parameters)
            stage_trigger_pulse_percent = self.parent.read_config_parameter('stage_trigger_pulse_%', self.cfg.asi_parameters)
            stage_delay_percent = self.parent.read_config_parameter('stage_trigger_delay_%', self.cfg.asi_parameters)
            stage_high_time = stage_trigger_pulse_percent * 0.01 * sweeptime
            stage_delay = stage_delay_percent * 0.01 * sweeptime
            self.stage_trigger_task.co_channels.add_co_pulse_chan_time(trig_line, high_time=stage_high_time, initial_delay=stage_delay)
            self.stage_trigger_task.triggers.start_trigger.cfg_dig_edge_start_trig(trig_source)

        '''Housekeeping: Setting up the AO task for the Galvo and setting the trigger input'''
        if self.ao_cards == 2:
            self.galvo_etl_task.ao_channels.add_ao_voltage_chan(ah['galvo_etl_task_line'])
            self.galvo_etl_task.timing.cfg_samp_clk_timing(rate=samplerate,
                                                       sample_mode=AcquisitionType.FINITE,
                                                       samps_per_chan=samples)
            self.galvo_etl_task.triggers.start_trigger.cfg_dig_edge_start_trig(ah['galvo_etl_task_trigger_source'])

            '''Housekeeping: Setting up the AO task for the ETL and lasers and setting the trigger input'''
            self.laser_task.ao_channels.add_ao_voltage_chan(ah['laser_task_line'])
            self.laser_task.timing.cfg_samp_clk_timing(rate=samplerate,
                                                        sample_mode=AcquisitionType.FINITE,
                                                        samps_per_chan=samples)
            self.laser_task.triggers.start_trigger.cfg_dig_edge_start_trig(ah['laser_task_trigger_source'])
        else:
            self.galvo_etl_laser_task.ao_channels.add_ao_voltage_chan(ah['galvo_etl_task_line'] + ',' + ah['laser_task_line'])
            self.galvo_etl_laser_task.timing.cfg_samp_clk_timing(rate=samplerate,
                                                       sample_mode=AcquisitionType.FINITE,
                                                       samps_per_chan=samples)
            self.galvo_etl_laser_task.triggers.start_trigger.cfg_dig_edge_start_trig(ah['galvo_etl_task_trigger_source'])

    def write_waveforms_to_tasks(self):
        """Write the waveforms to the slave tasks"""
        if self.ao_cards == 2:
            self.galvo_etl_task.write(self.galvo_and_etl_waveforms)
            self.laser_task.write(self.laser_waveforms)
        else:
            self.galvo_etl_laser_task.write(np.vstack((self.galvo_and_etl_waveforms, self.laser_waveforms)))

    def start_tasks(self):
        """Starts the tasks for camera triggering and analog outputs

        If the tasks are configured to be triggered, they won't output any
        signals until run_tasks() is called.
        """
        self.camera_trigger_task.start()
        if self.cfg.stage_parameters['stage_type'] in {'TigerASI'}:
            self.stage_trigger_task.start()
        if self.ao_cards == 2:
            self.galvo_etl_task.start()
            self.laser_task.start()
        else:
            self.galvo_etl_laser_task.start()

    def run_tasks(self):
        """Runs the tasks for triggering, analog and counter outputs

        Firstly, the master trigger triggers all other task via a shared trigger
        line (PFI line as given in the config file).

        For this to work, all analog output and counter tasks have to be started so
        that they are waiting for the trigger signal.
        """
        self.master_trigger_task.write([False, True, True, True, False], auto_start=True)

        '''Wait until everything is done - this is effectively a sleep function.'''
        if self.ao_cards == 2:
            self.galvo_etl_task.wait_until_done()
            self.laser_task.wait_until_done()
        else:
            self.galvo_etl_laser_task.wait_until_done()
        self.camera_trigger_task.wait_until_done()
        if self.cfg.stage_parameters['stage_type'] in {'TigerASI'}:
            self.stage_trigger_task.wait_until_done()

    def stop_tasks(self):
        """Stops the tasks for triggering, analog and counter outputs"""
        if self.ao_cards == 2:
            self.galvo_etl_task.stop()
            self.laser_task.stop()
        else:
            self.galvo_etl_laser_task.stop()
        self.camera_trigger_task.stop()
        if self.cfg.stage_parameters['stage_type'] in {'TigerASI'}:
            self.stage_trigger_task.stop()
        self.master_trigger_task.stop()

    def close_tasks(self):
        """Closes the tasks for triggering, analog and counter outputs.
        Tasks should only be closed are they are stopped.
        """
        if self.ao_cards == 2:
            self.galvo_etl_task.close()
            self.laser_task.close()
        else:
            self.galvo_etl_laser_task.close()
        self.camera_trigger_task.close()
        if self.cfg.stage_parameters['stage_type'] in {'TigerASI'}:
            self.stage_trigger_task.close()
        self.master_trigger_task.close()


class mesoSPIM_DemoWaveFormGenerator(mesoSPIM_WaveFormGenerator):
    """Demo subclass of mesoSPIM_WaveFormGenerator class
    """

    def __init__(self, parent):
        super().__init__(parent)

    def create_tasks(self):
        """"Demo version of the actual DAQmx-based function."""
        self.calculate_samples()
        samplerate, sweeptime = self.state.get_parameter_list(['samplerate','sweeptime'])
        camera_pulse_percent, camera_delay_percent = self.state.get_parameter_list(['camera_pulse_%','camera_delay_%'])
        self.camera_high_time = camera_pulse_percent*0.01*sweeptime
        self.camera_delay = camera_delay_percent*0.01*sweeptime

    def write_waveforms_to_tasks(self):
        """Demo: write the waveforms to the slave tasks """
        pass

    def start_tasks(self):
        """Demo: starts the tasks for camera triggering and analog outputs. """
        pass

    def run_tasks(self):
        """Demo: runs the tasks for triggering, analog and counter outputs. """
        time.sleep(self.state['sweeptime'])

    def stop_tasks(self):
        """"Demo: stop tasks"""
        pass

    def close_tasks(self):
        """Demo: closes the tasks for triggering, analog and counter outputs. """
        pass



#!/usr/bin/env python
"""
A ctypes based interface to Hamamatsu cameras.
(tested on a sCMOS Flash 4.0).
The documentation is a little confusing to me on this subject..
I used c_int32 when this is explicitly specified, otherwise I use c_int.
Hazen 10/13
George 11/17 - Updated for SDK4 and to allow fixed length acquisition

Updated for Win7/Win10 usage by Fabian Voigt, June 14th, 2018
"""

import ctypes
import ctypes.util
import numpy

# for debugging
import sys

# import storm_control.sc_library.halExceptions as halExceptions

# Hamamatsu constants.
DCAMCAP_EVENT_FRAMEREADY = int("0x0002", 0)
# DCAM4 API.
DCAMERR_ERROR = 0
DCAMERR_NOERROR = 1

DCAMPROP_ATTR_HASVALUETEXT = int("0x10000000", 0)
DCAMPROP_ATTR_READABLE = int("0x00010000", 0)
DCAMPROP_ATTR_WRITABLE = int("0x00020000", 0)

DCAMPROP_OPTION_NEAREST = int("0x80000000", 0)
DCAMPROP_OPTION_NEXT = int("0x01000000", 0)
DCAMPROP_OPTION_SUPPORT = int("0x00000000", 0)

DCAMPROP_TYPE_MODE = int("0x00000001", 0)
DCAMPROP_TYPE_LONG = int("0x00000002", 0)
DCAMPROP_TYPE_REAL = int("0x00000003", 0)
DCAMPROP_TYPE_MASK = int("0x0000000F", 0)

DCAMCAP_STATUS_ERROR = int("0x00000000", 0)
DCAMCAP_STATUS_BUSY = int("0x00000001", 0)
DCAMCAP_STATUS_READY = int("0x00000002", 0)
DCAMCAP_STATUS_STABLE = int("0x00000003", 0)
DCAMCAP_STATUS_UNSTABLE = int("0x00000004", 0)

DCAMWAIT_CAPEVENT_FRAMEREADY = int("0x0002", 0)
DCAMWAIT_CAPEVENT_STOPPED = int("0x0010", 0)

DCAMWAIT_RECEVENT_MISSED = int("0x00000200", 0)
DCAMWAIT_RECEVENT_STOPPED = int("0x00000400", 0)
DCAMWAIT_TIMEOUT_INFINITE = int("0x80000000", 0)

DCAM_DEFAULT_ARG = 0

DCAM_IDSTR_MODEL = int("0x04000104", 0)

DCAMCAP_TRANSFERKIND_FRAME = 0

DCAMCAP_START_SEQUENCE = -1
DCAMCAP_START_SNAP = 0

DCAMBUF_ATTACHKIND_FRAME = 0

# Hamamatsu structures.

## DCAMAPI_INIT
#
# The dcam initialization structure
#
class DCAMAPI_INIT(ctypes.Structure):
    _fields_ = [("size", ctypes.c_int32),
            ("iDeviceCount", ctypes.c_int32),
            ("reserved", ctypes.c_int32),
            ("initoptionbytes", ctypes.c_int32),
            ("initoption", ctypes.POINTER(ctypes.c_int32)),
            ("guid", ctypes.POINTER(ctypes.c_int32))]

## DCAMDEV_OPEN
#
# The dcam open structure
#
class DCAMDEV_OPEN(ctypes.Structure):
    _fields_ = [("size", ctypes.c_int32),
            ("index", ctypes.c_int32),
            ("hdcam", ctypes.c_void_p)]


## DCAMWAIT_OPEN
#
# The dcam wait open structure
#
class DCAMWAIT_OPEN(ctypes.Structure):
    _fields_ = [("size", ctypes.c_int32),
            ("supportevent", ctypes.c_int32),
            ("hwait", ctypes.c_void_p),
            ("hdcam", ctypes.c_void_p)]

## DCAMWAIT_START
#
# The dcam wait start structure
#
class DCAMWAIT_START(ctypes.Structure):
    _fields_ = [("size", ctypes.c_int32),
            ("eventhappened", ctypes.c_int32),
            ("eventmask", ctypes.c_int32),
            ("timeout", ctypes.c_int32)]

## DCAMCAP_TRANSFERINFO
#
# The dcam capture info structure
#
class DCAMCAP_TRANSFERINFO(ctypes.Structure):
    _fields_ = [("size", ctypes.c_int32),
            ("iKind", ctypes.c_int32),
            ("nNewestFrameIndex", ctypes.c_int32),
            ("nFrameCount", ctypes.c_int32)]


## DCAMBUF_ATTACH
#
# The dcam buffer attachment structure
#
class DCAMBUF_ATTACH(ctypes.Structure):
    _fields_ = [("size", ctypes.c_int32),
            ("iKind", ctypes.c_int32),
            ("buffer", ctypes.POINTER(ctypes.c_void_p)),
            ("buffercount", ctypes.c_int32)]

## DCAMBUF_FRAME
#
# The dcam buffer frame structure
#
class DCAMBUF_FRAME(ctypes.Structure):
    _fields_ = [("size", ctypes.c_int32),
            ("iKind", ctypes.c_int32),
            ("option", ctypes.c_int32),
            ("iFrame", ctypes.c_int32),
            ("buf", ctypes.c_void_p),
            ("rowbytes", ctypes.c_int32),
            ("type", ctypes.c_int32),
            ("width", ctypes.c_int32),
            ("height", ctypes.c_int32),
            ("left", ctypes.c_int32),
            ("top", ctypes.c_int32),
            ("timestamp", ctypes.c_int32),
            ("framestamp", ctypes.c_int32),
            ("camerastamp", ctypes.c_int32)]


## DCAMDEV_STRING
#
# The dcam device string structure
#
class DCAMDEV_STRING(ctypes.Structure):
    _fields_ = [("size", ctypes.c_int32),
            ("iString", ctypes.c_int32),
            ("text", ctypes.c_char_p),
            ("textbytes", ctypes.c_int32)]


## DCAMPROP_ATTR
#
# The dcam property attribute structure.
#
class DCAMPROP_ATTR(ctypes.Structure):
    _fields_ = [("cbSize", ctypes.c_int32),
                ("iProp", ctypes.c_int32),
                ("option", ctypes.c_int32),
                ("iReserved1", ctypes.c_int32),
                ("attribute", ctypes.c_int32),
                ("iGroup", ctypes.c_int32),
                ("iUnit", ctypes.c_int32),
                ("attribute2", ctypes.c_int32),
                ("valuemin", ctypes.c_double),
                ("valuemax", ctypes.c_double),
                ("valuestep", ctypes.c_double),
                ("valuedefault", ctypes.c_double),
                ("nMaxChannel", ctypes.c_int32),
                ("iReserved3", ctypes.c_int32),
                ("nMaxView", ctypes.c_int32),
                ("iProp_NumberOfElement", ctypes.c_int32),
                ("iProp_ArrayBase", ctypes.c_int32),
                ("iPropStep_Element", ctypes.c_int32)]

## DCAMPROP_VALUETEXT
#
# The dcam text property structure.
#
class DCAMPROP_VALUETEXT(ctypes.Structure):
    _fields_ = [("cbSize", ctypes.c_int32),
                ("iProp", ctypes.c_int32),
                ("value", ctypes.c_double),
                ("text", ctypes.c_char_p),
                ("textbytes", ctypes.c_int32)]


def convertPropertyName(p_name):
    """
    "Regularizes" a property name. We are using all lowercase names with
    the spaces replaced by underscores.
    """
    return p_name.lower().replace(" ", "_")


# class DCAMException():
#     pass


#
# Initialization
#
dcam = ctypes.windll.dcamapi

paraminit = DCAMAPI_INIT(0, 0, 0, 0, None, None)
paraminit.size = ctypes.sizeof(paraminit)
error_code = dcam.dcamapi_init(ctypes.byref(paraminit))
if (error_code != DCAMERR_NOERROR):
    raise DCAMException("DCAM initialization failed with error code " + str(error_code))

n_cameras = paraminit.iDeviceCount

class HCamData(object):
    """
    Hamamatsu camera data object.
    Initially I tried to use create_string_buffer() to allocate storage for the
    data from the camera but this turned out to be too slow. The software
    kept falling behind the camera and create_string_buffer() seemed to be the
    bottleneck.
    Using numpy makes a lot more sense anyways..
    """
    def __init__(self, size = None, **kwds):
        """
        Create a data object of the appropriate size.
        """
        super().__init__(**kwds)
        self.np_array = numpy.ascontiguousarray(numpy.empty(int(size/2), dtype=numpy.uint16))
        self.size = size

    def __getitem__(self, slice):
        return self.np_array[slice]

    def copyData(self, address):
        """
        Uses the C memmove function to copy data from an address in memory
        into memory allocated for the numpy array of this object.
        """
        ctypes.memmove(self.np_array.ctypes.data, address, self.size)

    def getData(self):
        return self.np_array

    def getDataPtr(self):
        return self.np_array.ctypes.data


class HamamatsuCamera(object):
    """
    Basic camera interface class.

    This version uses the Hamamatsu library to allocate camera buffers.
    Storage for the data from the camera is allocated dynamically and
    copied out of the camera buffers.
    """
    def __init__(self, camera_id = None, **kwds):
        """
        Open the connection to the camera specified by camera_id.
        """
        super().__init__(**kwds)

        self.buffer_index = 0
        self.camera_id = camera_id
        self.debug = False
        self.encoding = 'utf-8'
        self.frame_bytes = 0
        self.frame_x = 0
        self.frame_y = 0
        self.last_frame_number = 0
        self.properties = None
        self.max_backlog = 0
        self.number_image_buffers = 0

        self.acquisition_mode = "run_till_abort"
        self.number_frames = 0


        # Get camera model.
        self.camera_model = self.getModelInfo(camera_id)

        # Open the camera.
        paramopen = DCAMDEV_OPEN(0, self.camera_id, None)
        paramopen.size = ctypes.sizeof(paramopen)
        self.checkStatus(dcam.dcamdev_open(ctypes.byref(paramopen)),
                         "dcamdev_open")
        self.camera_handle = ctypes.c_void_p(paramopen.hdcam)

        # Set up wait handle
        paramwait = DCAMWAIT_OPEN(0, 0, None, self.camera_handle)
        paramwait.size = ctypes.sizeof(paramwait)
        self.checkStatus(dcam.dcamwait_open(ctypes.byref(paramwait)),
                "dcamwait_open")
        self.wait_handle = ctypes.c_void_p(paramwait.hwait)

        # Get camera properties.
        self.properties = self.getCameraProperties()

        # Get camera max width, height.
        # self.max_width = self.getPropertyValue("image_width")[0]
        # self.max_height = self.getPropertyValue("image_height")[0]
        self.max_width = self.getPropertyValue("image_width")
        self.max_height = self.getPropertyValue("image_height")


    def captureSetup(self):
        """
        Capture setup (internal use only). This is called at the start
        of new acquisition sequence to determine the current ROI and
        get the camera configured properly.
        """
        self.buffer_index = -1
        self.last_frame_number = 0

        # Set sub array mode.
        self.setSubArrayMode()

        # Get frame properties.
        self.frame_x = self.getPropertyValue("image_width")[0]
        self.frame_y = self.getPropertyValue("image_height")[0]
        self.frame_bytes = self.getPropertyValue("image_framebytes")[0]


    def checkStatus(self, fn_return, fn_name= "unknown"):
        """
        Check return value of the dcam function call.
        Throw an error if not as expected?
        """
        #if (fn_return != DCAMERR_NOERROR) and (fn_return != DCAMERR_ERROR):
        #    raise DCAMException("dcam error: " + fn_name + " returned " + str(fn_return))
        if (fn_return == DCAMERR_ERROR):
            c_buf_len = 80
            c_buf = ctypes.create_string_buffer(c_buf_len)
            c_error = dcam.dcam_getlasterror(self.camera_handle,
                                             c_buf,
                                             ctypes.c_int32(c_buf_len))
            raise DCAMException("dcam error " + str(fn_name) + " " + str(c_buf.value))
            #print "dcam error", fn_name, c_buf.value
        return fn_return

    def getCameraProperties(self):
        """
        Return the ids & names of all the properties that the camera supports. This
        is used at initialization to populate the self.properties attribute.
        """
        c_buf_len = 64
        c_buf = ctypes.create_string_buffer(c_buf_len)
        properties = {}
        prop_id = ctypes.c_int32(0)

        # Reset to the start.
        ret = dcam.dcamprop_getnextid(self.camera_handle,
                                      ctypes.byref(prop_id),
                                      ctypes.c_uint32(DCAMPROP_OPTION_NEAREST))
        if (ret != 0) and (ret != DCAMERR_NOERROR):
            self.checkStatus(ret, "dcamprop_getnextid")

        # Get the first property.
        ret = dcam.dcamprop_getnextid(self.camera_handle,
                                          ctypes.byref(prop_id),
                                          ctypes.c_int32(DCAMPROP_OPTION_NEXT))
        if (ret != 0) and (ret != DCAMERR_NOERROR):
            self.checkStatus(ret, "dcamprop_getnextid")
        self.checkStatus(dcam.dcamprop_getname(self.camera_handle,
                                                   prop_id,
                                                   c_buf,
                                                   ctypes.c_int32(c_buf_len)),
                         "dcamprop_getname")

        # Get the rest of the properties.
        last = -1
        while (prop_id.value != last):
            last = prop_id.value
            properties[convertPropertyName(c_buf.value.decode(self.encoding))] = prop_id.value
            ret = dcam.dcamprop_getnextid(self.camera_handle,
                                              ctypes.byref(prop_id),
                                              ctypes.c_int32(DCAMPROP_OPTION_NEXT))
            if (ret != 0) and (ret != DCAMERR_NOERROR):
                self.checkStatus(ret, "dcamprop_getnextid")
            self.checkStatus(dcam.dcamprop_getname(self.camera_handle,
                                                       prop_id,
                                                       c_buf,
                                                       ctypes.c_int32(c_buf_len)),
                             "dcamprop_getname")
        return properties

    def getFrames(self):
        """
        Gets all of the available frames.

        This will block waiting for new frames even if
        there new frames available when it is called.
        """
        frames = []
        for n in self.newFrames():

            paramlock = DCAMBUF_FRAME(
                    0, 0, 0, n, None, 0, 0, 0, 0, 0, 0, 0, 0, 0)
            paramlock.size = ctypes.sizeof(paramlock)

            # Lock the frame in the camera buffer & get address.
            self.checkStatus(dcam.dcambuf_lockframe(self.camera_handle,
                                                ctypes.byref(paramlock)),
                             "dcambuf_lockframe")

            # Create storage for the frame & copy into this storage.
            hc_data = HCamData(self.frame_bytes)
            hc_data.copyData(paramlock.buf)

            frames.append(hc_data)


        return [frames, [self.frame_x, self.frame_y]]

    def getModelInfo(self, camera_id):
        """
        Returns the model of the camera
        """

        c_buf_len = 20
        string_value = ctypes.create_string_buffer(c_buf_len)
        paramstring = DCAMDEV_STRING(
                        0,
                        DCAM_IDSTR_MODEL,
                        ctypes.cast(string_value, ctypes.c_char_p),
                        c_buf_len)
        paramstring.size = ctypes.sizeof(paramstring)

        self.checkStatus(dcam.dcamdev_getstring(ctypes.c_int32(camera_id),
                                                ctypes.byref(paramstring)),
                         "dcamdev_getstring")

        return string_value.value.decode(self.encoding)

    def getProperties(self):
        """
        Return the list of camera properties. This is the one to call if you
        want to know the camera properties.
        """
        return self.properties

    def getPropertyAttribute(self, property_name):
        """
        Return the attribute structure of a particular property.

        FIXME (OPTIMIZATION): Keep track of known attributes?
        """
        p_attr = DCAMPROP_ATTR()
        p_attr.cbSize = ctypes.sizeof(p_attr)
        p_attr.iProp = self.properties[property_name]
        ret = self.checkStatus(dcam.dcamprop_getattr(self.camera_handle,
                                                         ctypes.byref(p_attr)),
                               "dcamprop_getattr")
        if (ret == 0):
            print("property", property_id, "is not supported")
            return False
        else:
            return p_attr

    def getPropertyRange(self, property_name):
        """
        Return the range for an attribute.
        """
        prop_attr = self.getPropertyAttribute(property_name)
        temp = prop_attr.attribute & DCAMPROP_TYPE_MASK
        if (temp == DCAMPROP_TYPE_REAL):
            return [float(prop_attr.valuemin), float(prop_attr.valuemax)]
        else:
            return [int(prop_attr.valuemin), int(prop_attr.valuemax)]

    def getPropertyRW(self, property_name):
        """
        Return if a property is readable / writeable.
        """
        prop_attr = self.getPropertyAttribute(property_name)
        rw = []

        # Check if the property is readable.
        if (prop_attr.attribute & DCAMPROP_ATTR_READABLE):
            rw.append(True)
        else:
            rw.append(False)

        # Check if the property is writeable.
        if (prop_attr.attribute & DCAMPROP_ATTR_WRITABLE):
            rw.append(True)
        else:
            rw.append(False)

        return rw

    def getPropertyText(self, property_name):
        """
        #Return the text options of a property (if any).
        """
        prop_attr = self.getPropertyAttribute(property_name)
        if not (prop_attr.attribute & DCAMPROP_ATTR_HASVALUETEXT):
            return {}
        else:
            # Create property text structure.
            prop_id = self.properties[property_name]
            v = ctypes.c_double(prop_attr.valuemin)

            prop_text = DCAMPROP_VALUETEXT()
            c_buf_len = 64
            c_buf = ctypes.create_string_buffer(c_buf_len)
            #prop_text.text = ctypes.c_char_p(ctypes.addressof(c_buf))
            prop_text.cbSize = ctypes.c_int32(ctypes.sizeof(prop_text))
            prop_text.iProp = ctypes.c_int32(prop_id)
            prop_text.value = v
            prop_text.text = ctypes.addressof(c_buf)
            prop_text.textbytes = c_buf_len

            # Collect text options.
            done = False
            text_options = {}
            while not done:
                # Get text of current value.
                self.checkStatus(dcam.dcamprop_getvaluetext(self.camera_handle,
                                                ctypes.byref(prop_text)),
                                 "dcamprop_getvaluetext")
                text_options[prop_text.text.decode(self.encoding)] = int(v.value)

                # Get next value.
                ret = dcam.dcamprop_queryvalue(self.camera_handle,
                                           ctypes.c_int32(prop_id),
                                           ctypes.byref(v),
                                           ctypes.c_int32(DCAMPROP_OPTION_NEXT))
                prop_text.value = v

                if (ret != 1):
                    done = True

            return text_options

    def getPropertyValue(self, property_name):
        """
        Return the current setting of a particular property.
        """

        # Check if the property exists.
        if not (property_name in self.properties):
            print(" unknown property name:", property_name)
            return False
        prop_id = self.properties[property_name]

        # Get the property attributes.
        prop_attr = self.getPropertyAttribute(property_name)

        # Get the property value.
        c_value = ctypes.c_double(0)
        self.checkStatus(dcam.dcamprop_getvalue(self.camera_handle,
                                                    ctypes.c_int32(prop_id),
                                                    ctypes.byref(c_value)),
                         "dcamprop_getvalue")

        # Convert type based on attribute type.
        temp = prop_attr.attribute & DCAMPROP_TYPE_MASK
        if (temp == DCAMPROP_TYPE_MODE):
            prop_type = "MODE"
            prop_value = int(c_value.value)
        elif (temp == DCAMPROP_TYPE_LONG):
            prop_type = "LONG"
            prop_value = int(c_value.value)
        elif (temp == DCAMPROP_TYPE_REAL):
            prop_type = "REAL"
            prop_value = c_value.value
        else:
            prop_type = "NONE"
            prop_value = False

        return [prop_value, prop_type]

    def isCameraProperty(self, property_name):
        """
        Check if a property name is supported by the camera.
        """
        if (property_name in self.properties):
            return True
        else:
            return False

    def newFrames(self):
        """
        Return a list of the ids of all the new frames since the last check.
        Returns an empty list if the camera has already stopped and no frames
        are available.

        This will block waiting for at least one new frame.
        """
        # Wait for a new frame.
        # paramstart = DCAMWAIT_START(
        #         0, 0, DCAMCAP_EVENT_FRAMEREADY, DCAMWAIT_TIMEOUT_INFINITE)
        paramstart = DCAMWAIT_START(
                0, 0, DCAMCAP_EVENT_FRAMEREADY, 1000)

        paramstart.size = ctypes.sizeof(paramstart)

        self.checkStatus(dcam.dcamwait_start(self.wait_handle,
                                        ctypes.byref(paramstart)),
                         "dcamwait_start")
        # Removed and updated from an older version to get around a triggering bug
        # captureStatus = ctypes.c_int32(0)
        # self.checkStatus(dcam.dcamcap_status(
        #     self.camera_handle, ctypes.byref(captureStatus)))
        #
        # # Wait for a new frame if the camera is acquiring.
        # if captureStatus.value == DCAMCAP_STATUS_BUSY:
        #     paramstart = DCAMWAIT_START(
        #             0,
        #             0,
        #             DCAMWAIT_CAPEVENT_FRAMEREADY | DCAMWAIT_CAPEVENT_STOPPED,
        #             100)
        #     paramstart.size = ctypes.sizeof(paramstart)
        #     self.checkStatus(dcam.dcamwait_start(self.wait_handle,
        #                                     ctypes.byref(paramstart)),
        #                      "dcamwait_start")

        # Check how many new frames there are.
        paramtransfer = DCAMCAP_TRANSFERINFO(
                0, DCAMCAP_TRANSFERKIND_FRAME, 0, 0)
        paramtransfer.size = ctypes.sizeof(paramtransfer)
        self.checkStatus(dcam.dcamcap_transferinfo(self.camera_handle,
                                               ctypes.byref(paramtransfer)),
                         "dcamcap_transferinfo")
        cur_buffer_index = paramtransfer.nNewestFrameIndex
        cur_frame_number = paramtransfer.nFrameCount

        # Check that we have not acquired more frames than we can store in our buffer.
        # Keep track of the maximum backlog.
        backlog = cur_frame_number - self.last_frame_number
        if (backlog > self.number_image_buffers):
            print(">> Warning! hamamatsu camera frame buffer overrun detected!")
        if (backlog > self.max_backlog):
            self.max_backlog = backlog
        self.last_frame_number = cur_frame_number


        # Create a list of the new frames.
        new_frames = []
        if (cur_buffer_index < self.buffer_index):
            for i in range(self.buffer_index + 1, self.number_image_buffers):
                new_frames.append(i)
            for i in range(cur_buffer_index + 1):
                new_frames.append(i)
        else:
            for i in range(self.buffer_index, cur_buffer_index):
                new_frames.append(i+1)
        self.buffer_index = cur_buffer_index

        if self.debug:
            print(new_frames)

        return new_frames

    def setPropertyValue(self, property_name, property_value):
        """
        Set the value of a property.
        """

        # Check if the property exists.
        if not (property_name in self.properties):
            print(" unknown property name:", property_name)
            return False

        # If the value is text, figure out what the
        # corresponding numerical property value is.
        if (isinstance(property_value, str)):
            text_values = self.getPropertyText(property_name)
            if (property_value in text_values):
                property_value = float(text_values[property_value])
            else:
                print(" unknown property text value:", property_value, "for", property_name)
                return False

        # Check that the property is within range.
        [pv_min, pv_max] = self.getPropertyRange(property_name)
        if (property_value < pv_min):
            print(" set property value", property_value, "is less than minimum of", pv_min, property_name, "setting to minimum")
            property_value = pv_min
        if (property_value > pv_max):
            print(" set property value", property_value, "is greater than maximum of", pv_max, property_name, "setting to maximum")
            property_value = pv_max

        # Set the property value, return what it was set too.
        prop_id = self.properties[property_name]
        p_value = ctypes.c_double(property_value)
        self.checkStatus(dcam.dcamprop_setgetvalue(self.camera_handle,
                                           ctypes.c_int32(prop_id),
                                           ctypes.byref(p_value),
                                           ctypes.c_int32(DCAM_DEFAULT_ARG)),
                         "dcamprop_setgetvalue")
        return p_value.value

    def setSubArrayMode(self):
        """
        This sets the sub-array mode as appropriate based on the current ROI.
        """

        # Check ROI properties.
        roi_w = self.getPropertyValue("subarray_hsize")[0]
        roi_h = self.getPropertyValue("subarray_vsize")[0]

        # If the ROI is smaller than the entire frame turn on subarray mode
        if ((roi_w == self.max_width) and (roi_h == self.max_height)):
            self.setPropertyValue("subarray_mode", "OFF")
        else:
            self.setPropertyValue("subarray_mode", "ON")

    def setACQMode(self, mode, number_frames = None):
        '''
        Set the acquisition mode to either run until aborted or to
        stop after acquiring a set number of frames.
        mode should be either "fixed_length" or "run_till_abort"
        if mode is "fixed_length", then number_frames indicates the number
        of frames to acquire.
        '''

        self.stopAcquisition()

        if self.acquisition_mode is "fixed_length" or \
                self.acquisition_mode is "run_till_abort":
            self.acquisition_mode = mode
            self.number_frames = number_frames
        else:
            raise DCAMException("Unrecognized acqusition mode: " + mode)


    def startAcquisition(self):
        """
        Start data acquisition.
        """
        self.captureSetup()

        #
        # Allocate Hamamatsu image buffers.
        # We allocate enough to buffer 2 seconds of data or the specified
        # number of frames for a fixed length acquisition
        #
        if self.acquisition_mode is "run_till_abort":
            n_buffers = int(2.0*self.getPropertyValue("internal_frame_rate")[0])
        elif self.acquisition_mode is "fixed_length":
            n_buffers = self.number_frames

        self.number_image_buffers = n_buffers



        self.checkStatus(dcam.dcambuf_alloc(self.camera_handle,
                                  ctypes.c_int32(self.number_image_buffers)),
                         "dcambuf_alloc")

        # Start acquisition.
        if self.acquisition_mode is "run_till_abort":
            self.checkStatus(dcam.dcamcap_start(self.camera_handle,
                                    DCAMCAP_START_SEQUENCE),
                             "dcamcap_start")
        if self.acquisition_mode is "fixed_length":
            self.checkStatus(dcam.dcamcap_start(self.camera_handle,
                                    DCAMCAP_START_SNAP),
                             "dcamcap_start")

    def stopAcquisition(self):
        """
        Stop data acquisition.
        """

        # Stop acquisition.
        self.checkStatus(dcam.dcamcap_stop(self.camera_handle),
                         "dcamcap_stop")

        print("max camera backlog was", self.max_backlog, "of", self.number_image_buffers)
        self.max_backlog = 0

        # Free image buffers.
        self.number_image_buffers = 0
        self.checkStatus(dcam.dcambuf_release(self.camera_handle,
                                                DCAMBUF_ATTACHKIND_FRAME),
                         "dcambuf_release")

    def shutdown(self):
        """
        Close down the connection to the camera.
        """
        self.checkStatus(dcam.dcamwait_close(self.wait_handle),
                         "dcamwait_close")
        self.checkStatus(dcam.dcamdev_close(self.camera_handle),
                         "dcamdev_close")

    def sortedPropertyTextOptions(self, property_name):
        """
        Returns the property text options a list sorted by value.
        """
        text_values = self.getPropertyText(property_name)
        return sorted(text_values, key = text_values.get)


class HamamatsuCameraMR(HamamatsuCamera):
    """
    Memory recycling camera class.

    This version allocates "user memory" for the Hamamatsu camera
    buffers. This memory is also the location of the storage for
    the np_array element of a HCamData() class. The memory is
    allocated once at the beginning, then recycled. This means
    that there is a lot less memory allocation & shuffling compared
    to the basic class, which performs one allocation and (I believe)
    two copies for each frame that is acquired.

    WARNING: There is the potential here for chaos. Since the memory
             is now shared there is the possibility that downstream code
             will try and access the same bit of memory at the same time
             as the camera and this could end badly.
    FIXME: Use lockbits (and unlockbits) to avoid memory clashes?
           This would probably also involve some kind of reference
           counting scheme.
    """
    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.hcam_data = []
        self.hcam_ptr = False
        self.old_frame_bytes = -1

        self.setPropertyValue("output_trigger_kind[0]", 2)

    def getFrames(self):
        """
        Gets all of the available frames.

        This will block waiting for new frames even if there new frames
        available when it is called.

        FIXME: It does not always seem to block? The length of frames can
               be zero. Are frames getting dropped? Some sort of race condition?
        """
        frames = []
        for n in self.newFrames():
            frames.append(self.hcam_data[n])

        return [frames, [self.frame_x, self.frame_y]]

    def startAcquisition(self):
        """
        Allocate as many frames as will fit in 2GB of memory and start data acquisition.
        """
        self.captureSetup()

        # Allocate new image buffers if necessary. This will allocate
        # as many frames as can fit in 2GB of memory, or 2000 frames,
        # which ever is smaller. The problem is that if the frame size
        # is small than a lot of buffers can fit in 2GB. Assuming that
        # the camera maximum speed is something like 1KHz 2000 frames
        # should be enough for 2 seconds of storage, which will hopefully
        # be long enough.
        #
        if (self.old_frame_bytes != self.frame_bytes) or \
                (self.acquisition_mode is "fixed_length"):

            n_buffers = min(int((2.0 * 1024 * 1024 * 1024)/self.frame_bytes), 2000)
            if self.acquisition_mode is "fixed_length":
                self.number_image_buffers = self.number_frames
            else:
                self.number_image_buffers = n_buffers



            # Allocate new image buffers.
            ptr_array = ctypes.c_void_p * self.number_image_buffers
            self.hcam_ptr = ptr_array()
            self.hcam_data = []
            for i in range(self.number_image_buffers):
                hc_data = HCamData(self.frame_bytes)
                self.hcam_ptr[i] = hc_data.getDataPtr()
                self.hcam_data.append(hc_data)

            self.old_frame_bytes = self.frame_bytes

        # Attach image buffers and start acquisition.
        #
        # We need to attach & release for each acquisition otherwise
        # we'll get an error if we try to change the ROI in any way
        # between acquisitions.


        paramattach = DCAMBUF_ATTACH(0, DCAMBUF_ATTACHKIND_FRAME,
                self.hcam_ptr, self.number_image_buffers)
        paramattach.size = ctypes.sizeof(paramattach)

        if self.acquisition_mode is "run_till_abort":
            self.checkStatus(dcam.dcambuf_attach(self.camera_handle,
                                    paramattach),
                             "dcam_attachbuffer")
            self.checkStatus(dcam.dcamcap_start(self.camera_handle,
                                    DCAMCAP_START_SEQUENCE),
                             "dcamcap_start")
        if self.acquisition_mode is "fixed_length":
            paramattach.buffercount = self.number_frames
            self.checkStatus(dcam.dcambuf_attach(self.camera_handle,
                                    paramattach),
                             "dcambuf_attach")
            self.checkStatus(dcam.dcamcap_start(self.camera_handle,
                                    DCAMCAP_START_SNAP),
                             "dcamcap_start")


    def stopAcquisition(self):
        """
        Stop data acquisition and release the memory associates with the frames.
        """


        # Stop acquisition.
        self.checkStatus(dcam.dcamcap_stop(self.camera_handle),
                         "dcamcap_stop")

        # Release image buffers.
        if (self.hcam_ptr):
            self.checkStatus(dcam.dcambuf_release(self.camera_handle,
                                                DCAMBUF_ATTACHKIND_FRAME),
                         "dcambuf_release")

        ''' Print backlog only when it is large '''
        if self.max_backlog > 1:
            print("max camera backlog was:", self.max_backlog)
        self.max_backlog = 0


#
# Testing.
#
if (__name__ == "__main__"):

    import time
    import random

    print("found:", n_cameras, "cameras")
    if (n_cameras > 0):

        hcam = HamamatsuCameraMR(camera_id = 0)
        print(hcam.setPropertyValue("defect_correct_mode", 1))
        print("camera 0 model:", hcam.getModelInfo(0))

        # List support properties.
        if False:
            print("Supported properties:")
            props = hcam.getProperties()
            for i, id_name in enumerate(sorted(props.keys())):
                [p_value, p_type] = hcam.getPropertyValue(id_name)
                p_rw = hcam.getPropertyRW(id_name)
                read_write = ""
                if (p_rw[0]):
                    read_write += "read"
                if (p_rw[1]):
                    read_write += ", write"
                print("  ", i, ")", id_name, " = ", p_value, " type is:", p_type, ",", read_write)
                text_values = hcam.getPropertyText(id_name)
                if (len(text_values) > 0):
                    print("          option / value")
                    for key in sorted(text_values, key = text_values.get):
                        print("         ", key, "/", text_values[key])

        # Test setting & getting some parameters.
        if False:
            print(hcam.setPropertyValue("exposure_time", 0.001))

            #print(hcam.setPropertyValue("subarray_hsize", 2048))
            #print(hcam.setPropertyValue("subarray_vsize", 2048))
            print(hcam.setPropertyValue("subarray_hpos", 512))
            print(hcam.setPropertyValue("subarray_vpos", 512))
            print(hcam.setPropertyValue("subarray_hsize", 1024))
            print(hcam.setPropertyValue("subarray_vsize", 1024))

            print(hcam.setPropertyValue("binning", "1x1"))
            print(hcam.setPropertyValue("readout_speed", 2))

            hcam.setSubArrayMode()
            #hcam.startAcquisition()
            #hcam.stopAcquisition()

            params = ["internal_frame_rate",
                      "timing_readout_time",
                      "exposure_time"]

            #                      "image_height",
            #                      "image_width",
            #                      "image_framebytes",
            #                      "buffer_framebytes",
            #                      "buffer_rowbytes",
            #                      "buffer_top_offset_bytes",
            #                      "subarray_hsize",
            #                      "subarray_vsize",
            #                      "binning"]
            for param in params:
                print(param, hcam.getPropertyValue(param)[0])

        # Test 'run_till_abort' acquisition.
        if False:
            print("Testing run till abort acquisition")
            hcam.startAcquisition()
            cnt = 0
            for i in range(300):
                [frames, dims] = hcam.getFrames()
                for aframe in frames:
                    print(cnt, aframe[0:5])
                    cnt += 1

            print("Frames acquired: " + str(cnt))
            hcam.stopAcquisition()

        # Test 'fixed_length' acquisition.
        if True:
            for j in range (10):
                print("Testing fixed length acquisition")
                hcam.setACQMode("fixed_length", number_frames = 10)
                hcam.startAcquisition()
                cnt = 0
                iterations = 0
                while cnt < 11 and iterations < 20:
                    [frames, dims] = hcam.getFrames()
                    waitTime = random.random()*0.03
                    time.sleep(waitTime)
                    iterations += 1
                    print('Frames loaded: ' + str(len(frames)))
                    print('Wait time: ' + str(waitTime))
                    for aframe in frames:
                        print(cnt, aframe[0:5])
                        cnt += 1
                if cnt < 10:
                    print('##############Error: Not all frames found#########')
                    input("Press enter to continue")
                print("Frames acquired: " + str(cnt))
                hcam.stopAcquisition()

                hcam.setACQMode("run_till_abort")
                hcam.startAcquisition()
                time.sleep(random.random())
                contFrames = hcam.getFrames()
                hcam.stopAcquisition()



#
# The MIT License
#
# Copyright (c) 2013 Zhuang Lab, Harvard University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#



# To run the test:
# python -m test.test_serial
import unittest
import src.devices.filter_wheels.ludlcontrol as ludl
import src.devices.zoom.mesoSPIM_Zoom as zoomlib

"""
Filterwheel settings from config file
"""
filterwheel_parameters = {'filterwheel_type' : 'Ludl',
                          'COMport' : 'COM6'}
filterdict = {'Empty-Alignment' : 0,
              '405/50' : 1,
              '480/40' : 2,
              '525/50' : 3,
              '535/30' : 4,
              '590/50' : 5,
              '585/40' : 6,
              '405/488/561/640 m' : 7,
              }

'''
Zoom configuration from config file
'''
zoom_parameters = {'zoom_type' : 'Dynamixel',
                   'servo_id' :  1,
                   'COMport' : 'COM10',
                   'baudrate' : 1000000}

zoomdict = {'0.63x' : 3423,
            '0.8x' : 3071,
            '1x' : 2707,
            '1.25x' : 2389,
            '1.6x' : 2047,
            '2x' : 1706,
            '2.5x' : 1354,
            '3.2x' : 967,
            '4x' : 637,
            '5x' : 318,
            '6.3x' : 0}

class TestFilterWheel(unittest.TestCase):
    def setUp(self) -> None:
        """"This will be called for EVERY test method of the class."""
        if filterwheel_parameters['filterwheel_type'] == 'Ludl':
            self.fwheel = ludl.LudlFilterwheel(filterwheel_parameters['COMport'], filterdict)
        else:
            raise ValueError('Only Ludl filterwheel test is currently implemented')

    def test_multiple_filter_pos(self):
        n_cycles = 3
        for i_cycle in range(n_cycles):
            print(f"cycle {i_cycle}/{n_cycles}")
            for filter in filterdict.keys():
                self.fwheel.set_filter(filter, wait_until_done=True)
                print(f"filter {filter}")

    # def tearDown(self) -> None:
    #     """"Tidies up after EACH test method execution."""

class TestZoomServo(unittest.TestCase):
    def setUp(self) -> None:
        """"This will be called for EVERY test method of the class."""
        if zoom_parameters['zoom_type'] == 'Dynamixel':
            self.zoom = zoomlib.DynamixelZoom(zoomdict, zoom_parameters['COMport'],
                                              zoom_parameters['servo_id'], zoom_parameters['baudrate'])
        else:
            raise ValueError('Only Dynamixel zoom servo test is currently implemented')

    def test_multiple_zoom_pos(self):
        n_cycles = 3
        for i_cycle in range(n_cycles):
            print(f"cycle {i_cycle}/{n_cycles}")
            for zoomratio in zoomdict.keys():
                self.zoom.set_zoom(zoomratio, wait_until_done=True)
                print(f"zoom {zoomratio}")


if __name__ == '__main__':
    unittest.main()



'''
Serial thread for the mesoSPIM project
======================================

This thread handles all connections with serial devices such as stages,
filter wheels, zoom systems etc.
'''

import logging
logger = logging.getLogger(__name__)
from PyQt5 import QtCore

''' Import mesoSPIM modules '''
from .mesoSPIM_State import mesoSPIM_StateSingleton

from .mesoSPIM_FilterWheel import mesoSPIM_DemoFilterWheel, DynamixelFilterWheel, LudlFilterWheel, SutterLambda10BFilterWheel
from .mesoSPIM_Zoom import DynamixelZoom, DemoZoom
from .mesoSPIM_Stages import mesoSPIM_PI_1toN, mesoSPIM_PI_NtoN, mesoSPIM_ASI_Tiger_Stage, mesoSPIM_ASI_MS2000_Stage, mesoSPIM_DemoStage, mesoSPIM_GalilStages, mesoSPIM_PI_f_rot_and_Galil_xyz_Stages, mesoSPIM_PI_rot_and_Galil_xyzf_Stages, mesoSPIM_PI_rotz_and_Galil_xyf_Stages, mesoSPIM_PI_rotzf_and_Galil_xy_Stages


class mesoSPIM_Serial(QtCore.QObject):
    '''This class handles mesoSPIM serial connections'''
    sig_finished = QtCore.pyqtSignal()
    sig_state_request = QtCore.pyqtSignal(dict)
    sig_position = QtCore.pyqtSignal(dict)
    sig_zero_axes = QtCore.pyqtSignal(list)
    sig_unzero_axes = QtCore.pyqtSignal(list)
    sig_stop_movement = QtCore.pyqtSignal()
    sig_load_sample = QtCore.pyqtSignal()
    sig_unload_sample = QtCore.pyqtSignal()
    sig_mark_rotation_position = QtCore.pyqtSignal()
    sig_status_message = QtCore.pyqtSignal(str)
    sig_pause = QtCore.pyqtSignal(bool)
    
    def __init__(self, parent):
        super().__init__()

        ''' Assign the parent class to a instance variable for callbacks '''
        self.parent = parent
        self.cfg = parent.cfg
        self.state = mesoSPIM_StateSingleton()

        ''' Handling of state changing requests '''
        self.parent.sig_state_request.connect(self.state_request_handler)
        self.parent.sig_state_request_and_wait_until_done.connect(lambda sdict: self.state_request_handler(sdict, wait_until_done=True), type=3)

        ''' Attaching the filterwheel '''
        if self.cfg.filterwheel_parameters['filterwheel_type'] == 'Ludl':
            self.filterwheel = LudlFilterWheel(self.cfg.filterwheel_parameters['COMport'],self.cfg.filterdict)
        elif self.cfg.filterwheel_parameters['filterwheel_type'] == 'Dynamixel':
            self.filterwheel = DynamixelFilterWheel(self.cfg.filterdict, self.cfg.filterwheel_parameters['COMport'],
                                                    self.cfg.filterwheel_parameters['servo_id'],
                                                    self.cfg.filterwheel_parameters['baudrate'])
        elif self.cfg.filterwheel_parameters['filterwheel_type'] == 'Demo':
            self.filterwheel = mesoSPIM_DemoFilterWheel(self.cfg.filterdict)
        elif self.cfg.filterwheel_parameters['filterwheel_type'] == 'Sutter':
            self.filterwheel = SutterLambda10BFilterWheel(self.cfg.filterwheel_parameters['COMport'], self.cfg.filterdict)

        ''' Attaching the zoom '''
        if self.cfg.zoom_parameters['zoom_type'] == 'Dynamixel':
            self.zoom = DynamixelZoom(self.cfg.zoomdict, self.cfg.zoom_parameters['COMport'], self.cfg.zoom_parameters['servo_id'], self.cfg.zoom_parameters['baudrate'])
        elif self.cfg.zoom_parameters['zoom_type'] == 'DemoZoom':
            self.zoom = DemoZoom(self.cfg.zoomdict)

        ''' Attaching the stage '''
        if self.cfg.stage_parameters['stage_type'] in {'PI', 'PI_1controllerNstages'}:
            self.stage = mesoSPIM_PI_1toN(self)
        elif self.cfg.stage_parameters['stage_type'] == 'PI_NcontrollersNstages':
            self.stage = mesoSPIM_PI_NtoN(self)
        elif self.cfg.stage_parameters['stage_type'] == 'GalilStage':
            self.stage = mesoSPIM_GalilStages(self)
        elif self.cfg.stage_parameters['stage_type'] == 'PI_rot_and_Galil_xyzf':
            self.stage = mesoSPIM_PI_rot_and_Galil_xyzf_Stages(self)
        elif self.cfg.stage_parameters['stage_type'] == 'PI_f_rot_and_Galil_xyz':
            self.stage = mesoSPIM_PI_f_rot_and_Galil_xyz_Stages(self)
        elif self.cfg.stage_parameters['stage_type'] == 'PI_rotz_and_Galil_xyf':
            self.stage = mesoSPIM_PI_rotz_and_Galil_xyf_Stages(self)
        elif self.cfg.stage_parameters['stage_type'] == 'PI_rotzf_and_Galil_xy':
            self.stage = mesoSPIM_PI_rotzf_and_Galil_xy_Stages(self)
        elif self.cfg.stage_parameters['stage_type'] == 'TigerASI':
            self.stage = mesoSPIM_ASI_Tiger_Stage(self)
            self.stage.sig_pause.connect(self.pause)
            self.parent.sig_progress.connect(self.stage.log_slice)
        elif self.cfg.stage_parameters['stage_type'] == 'MS2000ASI':
            self.stage = mesoSPIM_ASI_MS2000_Stage(self)
            self.stage.sig_pause.connect(self.pause)
            self.parent.sig_progress.connect(self.stage.log_slice)
        elif self.cfg.stage_parameters['stage_type'] == 'DemoStage':
            self.stage = mesoSPIM_DemoStage(self)
        try:
            self.stage.sig_status_message.connect(self.send_status_message)
            self.stage.sig_position.connect(self.report_position)
        except:
            print('Stage not initalized! Please check the config file')

        ''' Wiring signals through to child objects '''
        self.parent.sig_move_relative.connect(self.move_relative)
        self.parent.sig_move_relative_and_wait_until_done.connect(lambda sdict: self.move_relative(sdict, wait_until_done=True), type=3)

        self.parent.sig_move_absolute.connect(self.move_absolute)
        self.parent.sig_move_absolute_and_wait_until_done.connect(lambda sdict: self.move_absolute(sdict, wait_until_done=True), type=3)

        self.parent.sig_zero_axes.connect(self.sig_zero_axes.emit)
        self.parent.sig_unzero_axes.connect(self.sig_unzero_axes.emit)
        self.parent.sig_stop_movement.connect(self.sig_stop_movement.emit)
        self.parent.sig_load_sample.connect(self.sig_load_sample.emit)
        self.parent.sig_unload_sample.connect(self.sig_unload_sample.emit)

        self.parent.sig_mark_rotation_position.connect(self.sig_mark_rotation_position.emit)
        self.parent.sig_go_to_rotation_position.connect(self.go_to_rotation_position)
        self.parent.sig_go_to_rotation_position_and_wait_until_done.connect(lambda: self.go_to_rotation_position(wait_until_done=True), type=3)

    @QtCore.pyqtSlot(dict)
    def state_request_handler(self, sdict, wait_until_done=False):
        for key, value in zip(sdict.keys(), sdict.values()):
            if key == 'filter':
                if wait_until_done:
                    self.set_filter(value, wait_until_done)
                else:
                    self.set_filter(value)
                logger.info(f'state change: {key}: {value}')
            if key == 'zoom':
                if wait_until_done:
                    self.set_zoom(value, wait_until_done)
                else:
                    self.set_zoom(value)
                logger.info(f'state change: {key}: {value}')
            if key == 'stage_program':
                self.execute_stage_program()
                logger.info(f'state change: {key}: {value}')
            if key == 'ttl_movement_enabled_during_acq':
                self.enable_ttl_motion(value)
                logger.info(f'state change: {key}: {value}')

    @QtCore.pyqtSlot(str)
    def send_status_message(self, string):
        self.sig_status_message.emit(string)

    @QtCore.pyqtSlot(bool)
    def pause(self, boolean):
        self.sig_pause.emit(boolean)

    @QtCore.pyqtSlot(bool)
    def enable_ttl_motion(self, boolean):
        self.stage.enable_ttl_motion(boolean)

    @QtCore.pyqtSlot(dict)
    def move_relative(self, sdict, wait_until_done=False):
        if wait_until_done:
            self.stage.move_relative(sdict, wait_until_done=True)
        else:
            self.stage.move_relative(sdict)

    @QtCore.pyqtSlot(dict)
    def move_absolute(self, sdict, wait_until_done=False):
        if wait_until_done:
            self.stage.move_absolute(sdict, wait_until_done=True)
        else:
            self.stage.move_absolute(sdict)

    @QtCore.pyqtSlot(dict)
    def report_position(self, sdict):
        self.sig_position.emit({'position': sdict})

    @QtCore.pyqtSlot()
    def go_to_rotation_position(self, wait_until_done=False):
        if wait_until_done:
            self.stage.go_to_rotation_position(wait_until_done=True)
        else:
            self.stage.go_to_rotation_position()

    @QtCore.pyqtSlot(str)
    def set_filter(self, sfilter, wait_until_done=False):
        # logger.info('Thread ID during set filter: '+str(int(QtCore.QThread.currentThreadId())))
        if wait_until_done:
            self.filterwheel.set_filter(sfilter, wait_until_done=True)
        else:
            self.filterwheel.set_filter(sfilter, wait_until_done=False)
        self.state['filter'] = sfilter

    @QtCore.pyqtSlot(str)
    def set_zoom(self, zoom, wait_until_done=False):
        # logger.info('Thread ID during set zoom: '+str(int(QtCore.QThread.currentThreadId())))
        ''' Here, the state parameters are set before sending the value to the zoom --
        this is to avoid laggy update loops with the GUI.'''
        self.state['zoom'] = zoom
        self.state['pixelsize'] = self.cfg.pixelsize[zoom]
        if wait_until_done:
            self.zoom.set_zoom(zoom, wait_until_done=True)
        else:
            self.zoom.set_zoom(zoom, wait_until_done=False)

    def execute_stage_program(self):
        self.stage.execute_program()



'''
Contains the joystick handlers

Because the signals emitted can only be processed when a QEventLoop is running, you
need something with an eventloop (e.g. a QApplication) even for testing.
'''
from PyQt5 import QtCore

from .Demo_SidePanel import Demo_SidePanel
from .logitech import FarmSimulatorSidePanel

class Demo_JoystickHandler(QtCore.QObject):
    def __init__(self, parent = None):
        super().__init__()
        # QtCore.QObject.__init__(self)

        self.parent = parent
        self.cfg = parent.cfg

    
class mesoSPIM_JoystickHandler(QtCore.QObject):

    def __init__(self, parent = None):
        super().__init__()
        # QtCore.QObject.__init__(self)

        self.parent = parent
        self.cfg = parent.cfg

        ''' parent is the window '''

        if self.cfg.sidepanel == 'FarmSimulator':
            self.joystick = FarmSimulatorSidePanel()

            self.joystick.sig_button_pressed.connect(self.button_handler)
            self.joystick.sig_mode_changed.connect(self.mode_handler)
            self.joystick.sig_axis_moved.connect(self.axis_handler)
        elif self.cfg.sidepanel == 'Demo':
            self.joystick = Demo_SidePanel()

        ''' '''
        self.SliderChangeCount = 0

    @QtCore.pyqtSlot(int)
    def button_handler(self, button_id):
        ''' Debugging print statement '''
        # print('Button pressed: ', button_id)

        ''' Laser switching buttons '''
        if button_id == 1:
            self.set_combobox_to_index(self.parent.LaserComboBox,0)
        if button_id == 2:
            self.set_combobox_to_index(self.parent.LaserComboBox,1)
        if button_id == 3:
            self.set_combobox_to_index(self.parent.LaserComboBox,2)
        if button_id == 6:
            self.set_combobox_to_index(self.parent.LaserComboBox,3)
        if button_id == 7:
            self.set_combobox_to_index(self.parent.LaserComboBox,4)
        if button_id == 8:
            self.set_combobox_to_index(self.parent.LaserComboBox,5)

        ''' Load & unload samples '''
        if button_id == 5:
            self.parent.sig_unload_sample.emit()
        if button_id == 10:
            self.parent.sig_load_sample.emit()

        ''' Filter & Zoom Increments & decrements '''
        if button_id == 11:
            self.increment_combobox(self.parent.FilterComboBox)
        if button_id == 12:
            self.decrement_combobox(self.parent.FilterComboBox)
        if button_id == 13:
            self.increment_combobox(self.parent.ZoomComboBox)
        if button_id == 14:
            self.decrement_combobox(self.parent.ZoomComboBox)

        ''' Shutter buttons '''
        if button_id == 17:
            self.set_combobox_to_string(self.parent.ShutterComboBox, 'Both')

        if button_id == 18:
            self.set_combobox_to_string(self.parent.ShutterComboBox, 'Both')

        if button_id == 19:
            self.set_combobox_to_string(self.parent.ShutterComboBox, 'Left')

        if button_id == 20:
            self.set_combobox_to_string(self.parent.ShutterComboBox, 'Right')

        ''' Live button '''
        if button_id == 21:
            current_state = self.parent.get_state_parameter('state')
            print('Current state: ',current_state)
            if current_state == ('live'):
                self.parent.StopButton.clicked.emit(True)
            elif current_state == 'idle':
                self.parent.LiveButton.clicked.emit(True)

        ''' Increase & decrease laser intensity '''
        if button_id == 26:
            self.increase_slider(self.parent.LaserIntensitySlider, 2)

        if button_id == 27:
            self.decrease_slider(self.parent.LaserIntensitySlider, 2)

        ''' Stop movement button '''
        if button_id == 28:
            self.parent.sig_stop_movement.emit()

        if button_id == 29:
            pass

    def set_combobox_to_index(self, combobox, index):
        if index < combobox.count():
            combobox.setCurrentIndex(index)

    def set_combobox_to_string(self, combobox, string):
        index = combobox.findText(string)
        ''' Debugging print statement '''
        # print('Index: ', index)
        if index != -1:
            combobox.setCurrentIndex(index)

    def increment_combobox(self, combobox):
        index = combobox.currentIndex()
        index += 1
        if index < combobox.count():
            combobox.setCurrentIndex(index)

    def decrement_combobox(self, combobox):
        index = combobox.currentIndex()
        index -= 1
        if index > -1:
            combobox.setCurrentIndex(index)

    def increase_slider(self, slider, event_devider=2):
        self.SliderChangeCount += 1
        ''' To avoid events coming too quickly,
        only every n-th event is causing a change if
        n = event_devider
        '''
        if self.SliderChangeCount % event_devider == 0:
            value = slider.value()
            value = value + 1

            if value != 100:
                slider.setValue(value)
            else:
                slider.setValue(100)

    def decrease_slider(self, slider, event_devider=2):
        ''' To avoid events coming too quickly,
        only every n-th event is causing a change if
        n = event_devider
        '''
        self.SliderChangeCount += 1

        if self.SliderChangeCount % event_devider == 0:
            value = slider.value()
            value = value - 1

            if value != 0:
                slider.setValue(value)
            else:
                slider.setValue(0)

    @QtCore.pyqtSlot(str)
    def mode_handler(self, str):
        '''
        Helper method to handle mode changes of the joystick.

        The FarmSimulatorSidePanel has 6 movement axes, 0 to 2 in "blue"
        LED mode (grey button on the joystick) and 3 to 5 in "red" mode.

        When starting up, the mode is unknown and has to be found out by
        registering which axes produce joystick events. 
        '''

        ''' Debugging print statement '''
        # print('New joystick mode: ', str)

        if str == '012':
            self.parent.display_status_message('Joystick Mode: XY Mode')
        elif str == '345':
            self.parent.display_status_message('Joystick Mode: ZF Mode')
        else:
            self.parent.display_status_message('Joystick Mode: Undefined')

    @QtCore.pyqtSlot(int, int)
    def axis_handler(self, axis_id, value):
        ''' The axis handler deals with joystick movements.

        The FarmSimulatorSidePanel has 6 movement axes, 0 to 2 in "blue"
        LED mode (grey button on the joystick) and 3 to 5 in "red" mode.

        When starting up, the mode is unknown and has to be found out by
        registering which axes produce joystick events.
        '''

        ''' Debugging print statement '''
        # print('Axis: ', axis_id, ',Value: ', value)

        ''' '''
        value = value - 128

        if axis_id == 0:
            self.parent.sig_move_relative.emit({'x_rel':value/5})
        elif axis_id == 1:
            self.parent.sig_move_relative.emit({'y_rel':value/5})
        elif axis_id == 3:
            ''' Some FarmSimulatorSidePanel have a bug which lets them
            send axis 2 and axis 3 (both rotation motions) at the same time.
            The following is intended to prevent this to affect the microscope:
            '''
            if self.joystick.mode == '123':
                pass
            else:
                self.parent.sig_move_relative.emit({'f_rel':value/30})
        elif axis_id == 4:
            self.parent.sig_move_relative.emit({'f_rel': value/5})
        elif axis_id == 5:
            self.parent.sig_move_relative.emit({'z_rel': value/5})



'''
Utilities for fast image-based optimization (auto-focus, etc)
Auto-focus is based on Autopilot paper (Royer at al, Nat Biotechnol. 2016 Dec;34(12):1267-1278. doi: 10.1038/nbt.3708.)
author: Nikita Vladimirov, @nvladimus, 2021
License: GPL-3
'''

from scipy.fftpack import dct
import numpy as np
import scipy.optimize

def _otf_radius(img, psf_radius_px):
    """Maximum number of spatial frequencies in the image"""
    assert len(img.shape) == 2, "Image must be 2D array"
    assert psf_radius_px > 0, "PSF radius must be positive"
    w = min(img.shape)
    psf_radius_px = np.ceil(psf_radius_px)  # clip all PSF radii below 1 px to 1.
    return w / (2 * psf_radius_px)

def _normL2(x):
    """L2 norm of n-dimensional array"""
    return np.sqrt(np.sum(x.flatten() ** 2))

def _abslog2(x):
    x_abs = abs(x)
    res = np.zeros(x_abs.shape)
    res[x_abs > 0] = np.log2(x_abs[x_abs > 0].astype(np.float64))
    return res

def _shannon(spectrum_2d, otf_radius=100):
    """Normalized shannon entropy of an image spectrum, bound by OTF support radius."""
    h, w = spectrum_2d.shape
    y, x = np.ogrid[:h, :w]
    support = (x + y < otf_radius)
    norm = _normL2(spectrum_2d[support])
    if norm != 0:
        terms = spectrum_2d[support].flatten() / norm
        entropy = -2 / otf_radius**2 * np.sum(abs(terms) * _abslog2(terms))
    else:
        entropy = 0
    return entropy

def _dct_2d(img, cutoff=100):
    cutoff = int(cutoff)
    assert len(img.shape) == 2, 'dct_2d(img): image must be 2D'
    return dct(dct(img.astype(np.float64).T, norm='ortho', n=cutoff).T, norm='ortho', n=cutoff)

def shannon_dct(img, psf_radius_px=1):
    """Shannon entropy of discreet cosine transform, for 2D images."""
    cutoff = _otf_radius(img, psf_radius_px)
    return _shannon(_dct_2d(img, cutoff), cutoff)

def gaussian_1d(x_arr, xo, sigma, amplitude=1, offset=0):
    """"Return 1D gaussian function as array"""
    g = offset + amplitude * np.exp(- ((x_arr - float(xo)) ** 2) / (2 * sigma ** 2))
    return g.ravel()

def _parabola_1d(x_arr, xcenter, amplitude=1, offset=0):
    """"Return 1D parabola function as array, with minus sign (e.g. reaching max at `xcenter`)"""
    g = offset - amplitude * (x_arr - float(xcenter)) ** 2
    return g.ravel()

def sigma2fwhm(sigma):
    """Convert gaussian std (sigma) to full-width half-maximum (FWHM)"""
    return 2.0 * sigma * np.sqrt(2 * np.log(2))

def _normalize_1d(arr, low_percentile=5.0):
    """Rescale the 1d array amplitude to [0,1] using percentile as a background value"""
    bg = np.percentile(arr, low_percentile)
    arr_normalized = np.clip((arr - bg) / (arr.max() - bg), 0, 1)
    return arr_normalized

def fit_gaussian_1d(f_arr, x_arr):
    """Fit measured values f_arr(x_arr) with a gaussian function.
    The true gaussian peak must lie withing the `x_arr` limits,
    so the measurements must 'clamp' the gaussian from left and right.

    Parameters:
    -----------
    f_arr: 1d-array
        Measured function values.
    x_arr: 1d-array
        Positions where the function was measured.

    Returns:
    --------
    (xcenter, sigma, f_amp, f_offset): tuple
        Parameters of gaussian fit: `f = f_offset + f_amp * np.exp(-((x_arr - xcenter)**2) / (2 * sigma ** 2))`
    """
    assert len(f_arr.shape) == len(x_arr.shape) == 1, f"Arrays must be 1d, got {f_arr.shape}, {x_arr.shape} instead"
    assert len(f_arr) >= 3, f"At least 3 points are needed for gaussian fit, received {len(f_arr)}."
    x_peak = x_arr[np.argmax(f_arr)]  # initial guess for gaussian peak position
    sigma_guess, amp_guess = x_arr.std(), f_arr.max()
    initial_guess = (x_peak, sigma_guess, amp_guess, amp_guess / 10)  # Parameters: xpos, sigma, amp, offset
    try:
        popt, pcov = scipy.optimize.curve_fit(gaussian_1d, (x_arr), f_arr,
                                              p0=initial_guess,
                                              bounds=((x_arr.min(), # min position of the peak
                                                       0.2 * sigma_guess,  # min sigma
                                                       0.5 * amp_guess, 0.001 * amp_guess),  # min amp, min offset
                                                      (x_arr.max(), # max position of the peak
                                                       20 * sigma_guess,  # max sigma
                                                       3 * amp_guess, 0.5 * amp_guess)))  # max amp, offset
    except RuntimeError as e:
        popt = initial_guess
        print(f"{e}")
    xcenter, sigma, f_amp, f_offset = popt
    return xcenter, sigma, f_amp, f_offset

def fit_parabola_1d(f_arr, x_arr):
    """Fit measured values f_arr(x_arr) with a parabola function f=a(x-x0)**2 + b.
    The true peak must lie withing the `x_arr` limits,
    so the measurements must 'clamp' the function from left and right.

    Parameters:
    -----------
    f_arr: 1d-array
        Measured function values.
    x_arr: 1d-array
        Positions where the function was measured.

    Returns:
    --------
    (xcenter, f_amp, f_offset): tuple
        Parameters of parabola fit: `f = f_offset + f_amp(x-xcenter)**2`
    """
    assert len(f_arr.shape) == len(x_arr.shape) == 1, f"Arrays must be 1d, got {f_arr.shape}, {x_arr.shape} instead"
    assert len(f_arr) >= 3, f"At least 3 points are needed for fit, received {len(f_arr)}."
    x_peak = x_arr[np.argmax(f_arr)]  # initial guess for peak position
    amp_guess, offset_guess = f_arr.max()/((x_arr.max() - x_arr.min()) / 2)**2, f_arr.max()
    initial_guess = (x_peak, amp_guess, offset_guess)  # Parameters: xpos, amp, offset
    try:
        popt, pcov = scipy.optimize.curve_fit(_parabola_1d, (x_arr), f_arr,
                                              p0=initial_guess,
                                              bounds=((x_arr.min(), # min position of the peak
                                                       0.001 * amp_guess, 0.001 * offset_guess),  # min amp, min offset
                                                      (x_arr.max(), # max position of the peak
                                                       1000 * amp_guess, 1000 * offset_guess)))  # max amp, offset
    except RuntimeError as e:
        popt = initial_guess
        print(f"{e}")
    xcenter,  f_amp, f_offset = popt
    return xcenter, f_amp, f_offset



'''
Core for the mesoSPIM project
=============================
'''
import os
import ctypes
import time
import platform
import io
import traceback

import logging
logger = logging.getLogger(__name__)

'''PyQt5 Imports'''
from PyQt5 import QtWidgets, QtCore, QtGui

'''National Instruments Imports'''
# import nidaqmx
# from nidaqmx.constants import AcquisitionType, TaskMode
# from nidaqmx.constants import LineGrouping, DigitalWidthUnits
# from nidaqmx.types import CtrTime

''' Import mesoSPIM modules '''
from .mesoSPIM_State import mesoSPIM_StateSingleton

from .devices.shutters.Demo_Shutter import Demo_Shutter
from .devices.shutters.NI_Shutter import NI_Shutter

from .mesoSPIM_Camera import mesoSPIM_Camera

from .devices.lasers.Demo_LaserEnabler import Demo_LaserEnabler
from .devices.lasers.mesoSPIM_LaserEnabler import mesoSPIM_LaserEnabler

from .mesoSPIM_Serial import mesoSPIM_Serial
# from .mesoSPIM_DemoSerial import mesoSPIM_Serial
from .mesoSPIM_WaveFormGenerator import mesoSPIM_WaveFormGenerator, mesoSPIM_DemoWaveFormGenerator

from .utils.acquisitions import AcquisitionList, Acquisition
from .utils.utility_functions import convert_seconds_to_string

class mesoSPIM_Core(QtCore.QObject):
    '''This class is the pacemaker of a mesoSPIM'''

    sig_finished = QtCore.pyqtSignal()

    sig_update_gui_from_state = QtCore.pyqtSignal(bool)

    sig_state_request = QtCore.pyqtSignal(dict)
    sig_state_request_and_wait_until_done = QtCore.pyqtSignal(dict)
    sig_position = QtCore.pyqtSignal(dict)

    sig_status_message = QtCore.pyqtSignal(str)
    sig_warning = QtCore.pyqtSignal(str)

    sig_progress = QtCore.pyqtSignal(dict)

    ''' Camera-related signals '''
    sig_prepare_image_series = QtCore.pyqtSignal(Acquisition, AcquisitionList)
    sig_add_images_to_image_series = QtCore.pyqtSignal(Acquisition, AcquisitionList)
    sig_add_images_to_image_series_and_wait_until_done = QtCore.pyqtSignal(Acquisition, AcquisitionList)
    sig_end_image_series = QtCore.pyqtSignal(Acquisition, AcquisitionList)
    sig_write_metadata = QtCore.pyqtSignal(Acquisition, AcquisitionList)

    sig_prepare_live = QtCore.pyqtSignal()
    sig_get_live_image = QtCore.pyqtSignal()
    sig_get_snap_image = QtCore.pyqtSignal(bool)
    sig_end_live = QtCore.pyqtSignal()

    ''' Movement-related signals: '''
    sig_move_relative = QtCore.pyqtSignal(dict)
    sig_move_relative_and_wait_until_done = QtCore.pyqtSignal(dict)
    sig_move_absolute = QtCore.pyqtSignal(dict)
    sig_move_absolute_and_wait_until_done = QtCore.pyqtSignal(dict)
    sig_zero_axes = QtCore.pyqtSignal(list)
    sig_unzero_axes = QtCore.pyqtSignal(list)
    sig_stop_movement = QtCore.pyqtSignal()
    sig_load_sample = QtCore.pyqtSignal()
    sig_unload_sample = QtCore.pyqtSignal()

    sig_mark_rotation_position = QtCore.pyqtSignal()
    sig_go_to_rotation_position = QtCore.pyqtSignal()
    sig_go_to_rotation_position_and_wait_until_done = QtCore.pyqtSignal()
    sig_polling_stage_position_start, sig_polling_stage_position_stop = QtCore.pyqtSignal(), QtCore.pyqtSignal()

    ''' ETL-related signals '''
    sig_save_etl_config = QtCore.pyqtSignal()

    def __init__(self, config, parent):
        super().__init__()

        ''' Assign the parent class to a instance variable for callbacks '''
        self.parent = parent
        self.package_directory = self.parent.package_directory
        self.cfg = self.parent.cfg

        self.state = mesoSPIM_StateSingleton()
        self.state['state'] = 'init'

        ''' The signal-slot switchboard '''
        # Note the name duplication (shadowing)!!
        # parent.sig_state_request -> self.state_request_handler
        # self.sig_state_request -> self.waveformer.state_request_handler
        self.parent.sig_state_request.connect(self.state_request_handler)

        self.parent.sig_execute_script.connect(self.execute_script)

        self.parent.sig_move_relative.connect(self.move_relative)
        # self.parent.sig_move_relative_and_wait_until_done.connect(lambda dict: self.move_relative(dict, wait_until_done=True))
        self.parent.sig_move_absolute.connect(self.move_absolute)
        # self.parent.sig_move_absolute_and_wait_until_done.connect(lambda dict: self.move_absolute(dict, wait_until_done=True))
        self.parent.sig_zero_axes.connect(self.zero_axes)
        self.parent.sig_unzero_axes.connect(self.unzero_axes)
        self.parent.sig_stop_movement.connect(self.stop_movement)
        self.parent.sig_load_sample.connect(self.sig_load_sample.emit)
        self.parent.sig_unload_sample.connect(self.sig_unload_sample.emit)
        self.parent.sig_save_etl_config.connect(self.sig_save_etl_config.emit)

        ''' Set the Camera thread up '''
        self.camera_thread = QtCore.QThread()
        self.camera_worker = mesoSPIM_Camera(self)
        #logger.info('Camera worker thread affinity before moveToThread? Answer:'+str(id(self.camera_worker.thread())))
        self.camera_worker.moveToThread(self.camera_thread)
        self.camera_worker.sig_update_gui_from_state.connect(self.sig_update_gui_from_state.emit)
        self.camera_worker.sig_status_message.connect(self.send_status_message_to_gui)
        self.camera_worker.sig_camera_frame.connect(self.parent.camera_window.set_image)
        #logger.info('Camera worker thread affinity after moveToThread? Answer:'+str(id(self.camera_worker.thread())))
        ''' Set the serial thread up '''
        #self.serial_thread = QtCore.QThread()
        self.serial_worker = mesoSPIM_Serial(self)
        # self.serial_worker.moveToThread(self.serial_thread)
        # If the stage (including the timer) is not manually moved to the serial thread, it will execute within the mesoSPIM_Core event loop - Fabian
        #self.serial_worker.stage.moveToThread(self.serial_thread)
        #self.serial_worker.stage.pos_timer.moveToThread(self.serial_thread)

        #self.serial_worker.sig_position.connect(lambda dict: self.sig_position.emit(dict))
        self.serial_worker.sig_position.connect(self.sig_position.emit)
        self.serial_worker.sig_status_message.connect(self.send_status_message_to_gui)
        self.serial_worker.sig_pause.connect(self.pause)
        self.sig_polling_stage_position_start.connect(self.serial_worker.stage.pos_timer.start)
        self.sig_polling_stage_position_stop.connect(self.serial_worker.stage.pos_timer.stop)

        ''' Start the threads '''
        self.camera_thread.start()
        #self.serial_thread.start()

        ''' Setting waveform generation up '''
        if self.cfg.waveformgeneration == 'NI':
            self.waveformer = mesoSPIM_WaveFormGenerator(self)
        elif self.cfg.waveformgeneration == 'DemoWaveFormGeneration':
            self.waveformer = mesoSPIM_DemoWaveFormGenerator(self)

        self.waveformer.sig_update_gui_from_state.connect(self.sig_update_gui_from_state.emit)
        self.sig_state_request.connect(self.waveformer.state_request_handler)
        self.sig_state_request_and_wait_until_done.connect(self.waveformer.state_request_handler)
        ''' If this line is activated while the waveformer and the core live in the same thread, a deadlock results '''
        #self.sig_state_request_and_wait_until_done.connect(self.waveformer.state_request_handler, type=3)

        ''' Setting the shutters up '''
        left_shutter_line = self.cfg.shutterdict['shutter_left']
        right_shutter_line = self.cfg.shutterdict['shutter_right']

        if self.cfg.shutter == 'NI':
            self.shutter_left = NI_Shutter(left_shutter_line)
            self.shutter_right = NI_Shutter(right_shutter_line)
        elif self.cfg.shutter == 'Demo':
            self.shutter_left = Demo_Shutter(left_shutter_line)
            self.shutter_right = Demo_Shutter(right_shutter_line)

        self.shutter_left.close()
        self.shutter_right.close()
        self.shutterswitch = self.cfg.shutterswitch if hasattr(self.cfg, 'shutterswitch') else False # backward compatibility with older config files
        self.state['shutterstate'] = False
        self.state['max_laser_voltage'] = self.cfg.startup['max_laser_voltage']

        ''' Setting the laser enabler up '''
        if self.cfg.laser == 'NI':
            self.laserenabler = mesoSPIM_LaserEnabler(self.cfg.laserdict)
        elif self.cfg.laser == 'Demo':
            self.laserenabler = Demo_LaserEnabler(self.cfg.laserdict)

        self.set_filter(self.cfg.startup['filter'])

        self.state['current_framerate'] = self.cfg.startup['average_frame_rate']
        self.state['snap_folder'] = self.cfg.startup['snap_folder']
        self.state['camera_display_live_subsampling'] = self.cfg.startup['camera_display_live_subsampling']
        self.state['camera_display_acquisition_subsampling'] = self.cfg.startup['camera_display_acquisition_subsampling']

        self.start_time = 0
        self.stopflag = False
        self.pauseflag = False

        if self.cfg.stage_parameters['stage_type'] in {'TigerASI'}:
            assert hasattr(self.cfg,  'asi_parameters'), "Config file for 'TigerASI' must contain 'asi_parameters' dict."
            self.TTL_mode_enabled_in_cfg = self.read_config_parameter('ttl_motion_enabled', self.cfg.asi_parameters)
        else:
            self.TTL_mode_enabled_in_cfg = False

        self.metadata_file = None
        # self.acquisition_list_rotation_position = {}
        self.state['state'] = 'idle'

    def __del__(self):
        '''Cleans the threads up after deletion, waits until the threads
        have truly finished their life.

        Make sure to keep this up to date with the number of threads
        '''
        try:
            self.camera_thread.quit()
            #self.serial_thread.quit()

            self.camera_thread.wait()
            #self.serial_thread.wait()
        except:
            pass


    @QtCore.pyqtSlot(dict)
    def state_request_handler(self, dict):
        for key, value in zip(dict.keys(),dict.values()):
            logger.info(f'State request: Key: {key}, Value: {value}')
            '''
            The request handling is done with exec() to write fewer lines of
            code.
            '''
            if key in ('filter',
                       'zoom',
                       'laser',
                       'intensity',
                       'shutterconfig',
                       'state',
                       'camera_exposure_time',
                       'camera_line_interval'):
                exec('self.set_'+key+'(value)')

            elif key in ('samplerate',
                       'sweeptime',
                       'ETL_cfg_file',
                       'etl_l_delay_%',
                       'etl_l_ramp_rising_%',
                       'etl_l_ramp_falling_%',
                       'etl_l_amplitude',
                       'etl_l_offset',
                       'etl_r_delay_%',
                       'etl_r_ramp_rising_%',
                       'etl_r_ramp_falling_%',
                       'etl_r_amplitude',
                       'etl_r_offset',
                       'galvo_l_frequency',
                       'galvo_l_amplitude',
                       'galvo_l_offset',
                       'galvo_l_duty_cycle',
                       'galvo_l_phase',
                       'galvo_r_frequency',
                       'galvo_r_amplitude',
                       'galvo_r_offset',
                       'galvo_r_duty_cycle',
                       'galvo_r_phase',
                       'laser_l_delay_%',
                       'laser_l_pulse_%',
                       'laser_l_max_amplitude',
                       'laser_r_delay_%',
                       'laser_r_pulse_%',
                       'laser_r_max_amplitude',
                       'camera_delay_%',
                       'camera_pulse_%',
                       'camera_display_live_subsampling',
                       'camera_display_acquisition_subsampling',
                       'camera_sensor_mode',
                       'camera_binning'
                       ):
                self.sig_state_request.emit({key : value})

    def set_state(self, state):
        if state == 'live':
            self.state['state'] = 'live'
            self.sig_state_request.emit({'state':'live'})
            logger.debug('Thread ID during live: '+str(int(QtCore.QThread.currentThreadId())))
            logger.debug('Core internal thread affinity in live: '+str(id(self.thread())))
            self.live()

        if state == 'snap':
            self.state['state']='snap'
            self.sig_state_request.emit({'state':'snap'})
            self.snap()

        elif state == 'run_selected_acquisition':
            self.state['state']= 'run_selected_acquisition'
            self.sig_state_request.emit({'state':'run_selected_acquisition'})
            self.start(row = self.state['selected_row'])

        elif state == 'run_acquisition_list':
            self.state['state'] = 'run_acquisition_list'
            self.sig_state_request.emit({'state':'run_acquisition_list'})
            self.start(row = None)

        elif state == 'preview_acquisition_with_z_update':
            self.state['state'] = 'preview_acquisition'
            self.preview_acquisition(z_update=True)

        elif state == 'preview_acquisition_without_z_update':
            self.state['state'] = 'preview_acquisition'
            self.preview_acquisition(z_update=False)

        elif state == 'idle':
            # print('Core: Stopping requested')
            self.sig_state_request.emit({'state':'idle'})
            self.stop()

        elif state == 'lightsheet_alignment_mode':
            self.state['state'] = 'lightsheet_alignment_mode'
            self.sig_state_request.emit({'state':'live'})
            self.lightsheet_alignment_mode()

        elif state == 'visual_mode':
            self.state['state'] = 'visual_mode'
            self.sig_state_request.emit({'state':'live'})
            self.visual_mode()

    def stop(self):
        self.stopflag = True
        ''' This stopflag is a bit risky, needs to be updated'''
        self.camera_worker.image_writer.abort_writing()
        self.state['state'] = 'idle'
        self.sig_update_gui_from_state.emit(False)
        self.sig_finished.emit()

    @QtCore.pyqtSlot(bool)
    def pause(self, boolean):
        self.pauseflag = boolean

    def send_progress(self,
                      cur_acq,
                      tot_acqs,
                      cur_image,
                      images_in_acq,
                      total_image_count,
                      image_counter,
                      time_passed_string,
                      remaining_time_string):

        dict = {'current_acq':cur_acq,
                'total_acqs' :tot_acqs,
                'current_image_in_acq':cur_image,
                'images_in_acq': images_in_acq,
                'total_image_count':total_image_count,
                'image_counter':image_counter,
                'time_passed_string': time_passed_string,
                'remaining_time_string': remaining_time_string,
        }
        self.sig_progress.emit(dict)

    @QtCore.pyqtSlot(dict)
    def set_filter(self, filter, wait_until_done=False):
        if wait_until_done:
            self.sig_state_request_and_wait_until_done.emit({'filter' : filter})
        else:
            self.sig_state_request.emit({'filter' : filter})

    @QtCore.pyqtSlot(dict)
    def set_zoom(self, zoom, wait_until_done=False, update_etl=True):
        if wait_until_done:
            self.sig_state_request_and_wait_until_done.emit({'zoom' : zoom})
            if update_etl:
                self.sig_state_request_and_wait_until_done.emit({'set_etls_according_to_zoom' : zoom})
        else:
            self.sig_state_request.emit({'zoom' : zoom})
            if update_etl:
                self.sig_state_request.emit({'set_etls_according_to_zoom' : zoom})

    @QtCore.pyqtSlot(str)
    def set_laser(self, laser, wait_until_done=False, update_etl=True):
        if wait_until_done:
            self.sig_state_request_and_wait_until_done.emit({'laser' : laser})
            if update_etl:
                self.sig_state_request_and_wait_until_done.emit({'set_etls_according_to_laser' : laser})
        else:
            self.sig_state_request.emit({'laser':laser})
            if update_etl:
                #print("ETL updated")
                self.sig_state_request.emit({'set_etls_according_to_laser' : laser})

    @QtCore.pyqtSlot(str)
    def set_intensity(self, intensity, wait_until_done=False):
        if wait_until_done:
            self.sig_state_request_and_wait_until_done.emit({'intensity': intensity})
        else:
            self.sig_state_request.emit({'intensity':intensity})

    @QtCore.pyqtSlot(float)
    def set_camera_exposure_time(self, time):
        self.sig_state_request.emit({'camera_exposure_time' : time})

    @QtCore.pyqtSlot(float)
    def set_camera_line_interval(self, time):
        self.sig_state_request.emit({'camera_line_interval' : time})

    @QtCore.pyqtSlot(dict)
    def move_relative(self, dict, wait_until_done=False):
        if wait_until_done:
            self.sig_move_relative_and_wait_until_done.emit(dict)
        else:
            self.sig_move_relative.emit(dict)

    @QtCore.pyqtSlot(dict)
    def move_absolute(self, dict, wait_until_done=False):
        if wait_until_done:
            self.sig_move_absolute_and_wait_until_done.emit(dict)
        else:
            self.sig_move_absolute.emit(dict)

    @QtCore.pyqtSlot(list)
    def zero_axes(self, list):
        self.sig_zero_axes.emit(list)

    @QtCore.pyqtSlot(list)
    def unzero_axes(self, list):
        self.sig_unzero_axes.emit(list)

    @QtCore.pyqtSlot()
    def stop_movement(self):
        self.sig_stop_movement.emit()

    @QtCore.pyqtSlot(str)
    def set_shutterconfig(self, shutterconfig):
        self.sig_state_request.emit({'shutterconfig': shutterconfig})

    @QtCore.pyqtSlot()
    def open_shutters(self):
        '''Here the left/right mode is hacked in
        If shutterswitch = True in the config:
        Assumes that the shutter_left line is the general shutter 
        and the shutter_right line is the left/right switch (Right==True)
        '''
        shutterconfig = self.state['shutterconfig']
        if shutterconfig == 'Both':
            if self.shutterswitch is False:
                self.shutter_left.open()
                self.shutter_right.open()
        elif shutterconfig == 'Left':
            if self.shutterswitch is False:
                self.shutter_left.open()
                self.shutter_right.close()
            else:
                self.shutter_left.open() # open the general shutter
                self.shutter_right.close() # set side-switch to false 

        elif shutterconfig == 'Right':
            if self.shutterswitch is False:
                self.shutter_right.open()
                self.shutter_left.close()
            else:
                self.shutter_left.open() # open the general shutte
                self.shutter_right.open() # set side-switch to true
        else: # BOTH open
            self.shutter_right.open()
            self.shutter_left.open()

        self.state['shutterstate'] = True

    @QtCore.pyqtSlot()
    def close_shutters(self):
        '''
        If shutterswitch = True in the config:
        Assumes that the shutter_left line is the general shutter 
        and the shutter_right line is the left/right switch (Right==True)
        '''
        if self.shutterswitch is False:
            self.shutter_left.close()
            self.shutter_right.close()
        else:
            self.shutter_left.close()
        
        self.state['shutterstate'] = False

    '''
    Sub-Imaging modes
    '''
    def snap(self, write_flag=True, laser_blanking=True):
        self.sig_prepare_live.emit()
        self.open_shutters()
        self.snap_image(laser_blanking)
        self.sig_get_snap_image.emit(write_flag)
        self.close_shutters()
        self.sig_end_live.emit()
        self.sig_finished.emit()
        QtWidgets.QApplication.processEvents()

    def snap_image(self, laser_blanking=True):
        '''Snaps a single image after updating the waveforms.

        Can be used in acquisitions where changing waveforms are required,
        but there is additional overhead due to the need to write the
        waveforms into the buffers of the NI cards.
        '''
        self.waveformer.create_tasks()
        self.waveformer.write_waveforms_to_tasks()
        laser = self.state['laser']
        if laser_blanking:
            self.laserenabler.enable(laser)
        self.waveformer.start_tasks()
        self.waveformer.run_tasks()
        if laser_blanking:
            self.laserenabler.disable_all()
        self.waveformer.stop_tasks()
        self.waveformer.close_tasks()

    def prepare_image_series(self):
        '''Prepares an image series without waveform update'''
        self.waveformer.create_tasks()
        self.waveformer.write_waveforms_to_tasks()

    def snap_image_in_series(self, laser_blanking=True):
        '''Snaps and image from a series without waveform update'''
        laser = self.state['laser']
        if laser_blanking:
            self.laserenabler.enable(laser)
        self.waveformer.start_tasks()
        self.waveformer.run_tasks()
        self.waveformer.stop_tasks()
        if laser_blanking:
            self.laserenabler.disable_all()

    def close_image_series(self):
        '''Cleans up after series without waveform update'''
        self.waveformer.close_tasks()
        logger.info("close_image_series() finished")

    def live(self):
        self.stopflag = False
        self.sig_prepare_live.emit()
        laser = self.state['laser']
        laser_blanking = False if (hasattr(self.cfg, 'laser_blanking') and (self.cfg.laser_blanking in ('stack', 'stacks'))) else True
        self.laserenabler.enable(laser)
        while self.stopflag is False:
            ''' How to handle a possible shutter switch?'''
            self.open_shutters()
            ''' Needs update to use snap image in series '''
            self.snap_image(laser_blanking)
            self.sig_get_live_image.emit()

            while self.pauseflag is True:
                time.sleep(0.1)
                QtWidgets.QApplication.processEvents()

            QtWidgets.QApplication.processEvents()

        self.laserenabler.disable_all()
        self.close_shutters()
        self.sig_end_live.emit()
        self.sig_finished.emit()

    def start(self, row=None):
        self.stopflag = False
        if row is None:
            acq_list = self.state['acq_list']
        else:
            acquisition = self.state['acq_list'][row]
            acq_list = AcquisitionList([acquisition])
            
        nonexisting_folders_list = acq_list.check_for_nonexisting_folders()
        filename_list = acq_list.check_for_existing_filenames()
        duplicates_list = acq_list.check_for_duplicated_filenames()
        files_without_extensions = acq_list.check_filename_extensions()
        free_disk_space_bytes = self.get_free_disk_space(acq_list)
        total_required_bytes = self.get_required_disk_space(acq_list)

        if nonexisting_folders_list:
            self.sig_warning.emit('The following folders do not exist - stopping! \n'+self.list_to_string_with_carriage_return(nonexisting_folders_list))
            self.sig_finished.emit()
        elif filename_list:
            self.sig_warning.emit('The following files already exist - stopping! \n'+self.list_to_string_with_carriage_return(filename_list))
            self.sig_finished.emit()
        elif duplicates_list:
            self.sig_warning.emit('The following filenames are duplicated - stopping! \n' +self.list_to_string_with_carriage_return(duplicates_list))
            self.sig_finished.emit()
        elif files_without_extensions:
            self.sig_warning.emit('Some files have no extensions (.raw, .tiff, .h5) - stopping! \n' + self.list_to_string_with_carriage_return(files_without_extensions))
            self.sig_finished.emit()
        elif free_disk_space_bytes < total_required_bytes * 1.1:
            self.sig_warning.emit('Not sufficient disk space - stopping! \n')
            self.sig_finished.emit()
        else:
            self.sig_update_gui_from_state.emit(True)
            self.prepare_acquisition_list(acq_list)
            self.run_acquisition_list(acq_list)
            self.close_acquisition_list(acq_list)
            self.sig_update_gui_from_state.emit(False)

    def formatSize(self, bytes):
        try:
            bytes = float(bytes)
            kb = bytes / 1024
        except Exception as e:
            print(f"{e}")
            return None
        if kb >= 1024:
            M = kb / 1024
            if M >= 1024:
                G = M / 1024
                return "%.2fG" % (G)
            else:
                return "%.2fM" % (M)
        else:
            return "%.2fkb" % (kb)

    def get_free_disk_space(self, acq_list):
        """Take the disk location of the first file and compute the free disk space"""
        filename0 = os.path.realpath(acq_list.get_all_filenames()[0])
        disk_name = os.path.splitdrive(filename0)[0]
        if platform.system() == 'Windows':
            free_bytes = ctypes.c_ulonglong(0)
            ctypes.windll.kernel32.GetDiskFreeSpaceExW(ctypes.c_wchar_p(disk_name), None, None, ctypes.pointer(free_bytes))
            print(f"Free disk {disk_name} space {self.formatSize(free_bytes.value)}")
            return free_bytes.value
        else: # non-Windows case, untested!
            st = os.statvfs(disk_name)
            return st.f_bavail * st.f_frsize

    def get_required_disk_space(self, acq_list):
        """"Compute total image data size from the acquisition list, in bytes"""
        BYTES_PER_PIXEL = 2 # 16-bit camera
        px_per_image = self.camera_worker.x_pixels * self.camera_worker.y_pixels
        total_bytes_required = acq_list.get_image_count() * px_per_image * BYTES_PER_PIXEL
        print(f"Required disk space: {self.formatSize(total_bytes_required)}")
        return total_bytes_required

    def prepare_acquisition_list(self, acq_list):
        ''' Housekeeping: Prepare the acquisition list '''
        self.image_count = 0
        self.acquisition_count = 0
        self.total_acquisition_count = len(acq_list)
        self.total_image_count = acq_list.get_image_count()
        self.start_time = time.time()

    def run_acquisition_list(self, acq_list):
        for acq in acq_list:
            if not self.stopflag:
                self.prepare_acquisition(acq, acq_list)
                self.run_acquisition(acq, acq_list)
                self.close_acquisition(acq, acq_list)

    def close_acquisition_list(self, acq_list):
        self.sig_status_message.emit('Closing Acquisition List')
        if not self.stopflag:
            current_rotation = self.state['position']['theta_pos']
            startpoint = acq_list.get_startpoint()
            target_rotation = startpoint['theta_abs']

            if current_rotation > target_rotation+0.1 or current_rotation < target_rotation-0.1:
                self.sig_go_to_rotation_position_and_wait_until_done.emit()
                self.move_absolute({'theta_abs':target_rotation}, wait_until_done=True)

            self.state['state'] = 'idle'
            self.move_absolute(acq_list.get_startpoint())

            self.set_filter(acq_list[0]['filter'])
            self.set_laser(acq_list[0]['laser'], wait_until_done=False, update_etl=False)
            self.set_zoom(acq_list[0]['zoom'], wait_until_done=False, update_etl=False)
            ''' This is for the GUI to update properly, otherwise ETL values for previous laser might be displayed '''
            QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 1)

            self.sig_state_request.emit({'etl_l_amplitude' : acq_list[0]['etl_l_amplitude']})
            self.sig_state_request.emit({'etl_r_amplitude' : acq_list[0]['etl_r_amplitude']})
            self.sig_state_request.emit({'etl_l_offset' : acq_list[0]['etl_l_offset']})
            self.sig_state_request.emit({'etl_r_offset' : acq_list[0]['etl_r_offset']})
            self.set_intensity(acq_list[0]['intensity'])
            time.sleep(0.1) # tiny sleep period to allow Main Window indicators to catch up
            self.sig_finished.emit()

    def preview_acquisition(self, z_update=True):
        self.stopflag = False
        row = self.state['selected_row']
        #self.sig_update_gui_from_state.emit(True) #deprecated
        acq = self.state['acq_list'][row]

        ''' Rotation handling goes here '''
        current_rotation = self.state['position']['theta_pos']
        startpoint = acq.get_startpoint()
        target_rotation = startpoint['theta_abs']

        ''' Create a flag when rotation is required: '''
        if current_rotation > target_rotation+0.1 or current_rotation < target_rotation-0.1:
            rotationflag = True
        else:
            rotationflag = False

        ''' Remove z-coordinate from dict so that z is not updated during preview: '''
        if not z_update:
            ''' If a rotation is necessary, z will be updated '''
            if not rotationflag:
                del startpoint['z_abs']

        ''' Check if sample has to be rotated, allow some tolerance '''
        if rotationflag:
            self.sig_status_message.emit('Going to rotation position')
            self.sig_go_to_rotation_position_and_wait_until_done.emit()
            self.sig_status_message.emit('Rotating sample')
            self.move_absolute({'theta_abs':target_rotation}, wait_until_done=True)

        self.sig_status_message.emit('Setting Filter')
        self.set_filter(acq['filter'], wait_until_done=True)

        self.sig_status_message.emit('Going to start position')
        self.move_absolute(startpoint, wait_until_done=False)

        self.sig_status_message.emit('Setting Shutter')
        self.set_shutterconfig(acq['shutterconfig'])
        self.sig_status_message.emit('Setting Zoom & Laser')
        self.set_zoom(acq['zoom'], wait_until_done=False, update_etl=False)
        self.set_intensity(acq['intensity'], wait_until_done=True)
        self.set_laser(acq['laser'], wait_until_done=True, update_etl=False)
        # Deprecated: This was for the GUI to update properly, otherwise ETL values for previous laser might be displayed
        #QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 1)

        self.sig_state_request.emit({'etl_l_amplitude' : acq['etl_l_amplitude']})
        self.sig_state_request.emit({'etl_r_amplitude' : acq['etl_r_amplitude']})
        self.sig_state_request.emit({'etl_l_offset' : acq['etl_l_offset']})
        self.sig_state_request.emit({'etl_r_offset' : acq['etl_r_offset']})

        self.sig_status_message.emit('Ready for preview...')
        self.sig_update_gui_from_state.emit(False)

        self.state['state'] = 'idle'

    def prepare_acquisition(self, acq, acq_list):
        ''' Housekeeping: Prepare the acquisition  '''
        logger.info(f'Core: Running Acquisition #{self.acquisition_count} with Filename: {acq["filename"]}')
        self.sig_status_message.emit('Going to start position')
        current_rotation = self.state['position']['theta_pos']
        startpoint = acq.get_startpoint()
        target_rotation = startpoint['theta_abs']
        self.acq_start_time_string = time.strftime("%Y%m%d-%H%M%S")

        self.sig_status_message.emit('Going to start position')
        ''' Check if sample has to be rotated, allow some tolerance '''
        if current_rotation > target_rotation+0.1 or current_rotation < target_rotation-0.1:
            self.sig_go_to_rotation_position_and_wait_until_done.emit()
            self.move_absolute({'theta_abs':target_rotation}, wait_until_done=True)
        
        self.move_absolute(startpoint, wait_until_done=True)
        self.sig_status_message.emit('Setting Filter & Shutter')
        self.set_shutterconfig(acq['shutterconfig'])
        self.set_filter(acq['filter'], wait_until_done=True)
        self.sig_status_message.emit('Setting Zoom')
        self.set_zoom(acq['zoom'], wait_until_done=False, update_etl=False)
        self.set_intensity(acq['intensity'], wait_until_done=True)
        self.set_laser(acq['laser'], wait_until_done=True, update_etl=False)
        ''' This is for the GUI to update properly, otherwise ETL values for previous laser might be displayed '''
        QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 1)

        self.sig_state_request.emit({'etl_l_amplitude' : acq['etl_l_amplitude']})
        self.sig_state_request.emit({'etl_r_amplitude' : acq['etl_r_amplitude']})
        self.sig_state_request.emit({'etl_l_offset' : acq['etl_l_offset']})
        self.sig_state_request.emit({'etl_r_offset' : acq['etl_r_offset']})
        self.f_step_generator = acq.get_focus_stepsize_generator()

        if self.TTL_mode_enabled_in_cfg is True:
            ''' The relative movement has to be carried out once with the ASI-controller '''
            self.move_relative(acq.get_delta_z_and_delta_f_dict(inverted=True))
            time.sleep(0.1)
            self.move_relative(acq.get_delta_z_and_delta_f_dict())
            time.sleep(0.1)
            self.sig_state_request.emit({'ttl_movement_enabled_during_acq' : True})
            time.sleep(0.05)

        # stop asking stages about their positions, to avoid messing up serial comm during acquisition:
        self.sig_polling_stage_position_stop.emit()

        self.sig_status_message.emit('Preparing camera: Allocating memory')
        self.sig_prepare_image_series.emit(acq, acq_list)
        self.prepare_image_series()
        self.sig_write_metadata.emit(acq, acq_list)

    def run_acquisition(self, acq, acq_list):
        steps = acq.get_image_count()
        self.sig_status_message.emit('Running Acquisition')
        self.open_shutters()
        self.image_acq_start_time = time.time()
        self.image_acq_start_time_string = time.strftime("%Y%m%d-%H%M%S")

        move_dict = acq.get_delta_dict()
        laser = self.state['laser']
        self.laserenabler.enable(laser)
        laser_blanking = False if (hasattr(self.cfg, 'laser_blanking') and (self.cfg.laser_blanking in ('stack', 'stacks'))) else True
        for i in range(steps):
            if self.stopflag is True:
                self.close_image_series()
                self.sig_end_image_series.emit(acq, acq_list)
                self.sig_finished.emit()
                break
            else:
                self.snap_image_in_series(laser_blanking)
                self.sig_add_images_to_image_series.emit(acq, acq_list)
                #time.sleep(0.02)
                # self.sig_add_images_to_image_series_and_wait_until_done.emit()

                # self.move_relative(acq.get_delta_z_dict(), wait_until_done=True)

                ''' Get the current correct f_step'''
                f_step = self.f_step_generator.__next__()
                if f_step != 0:
                    # print('F step: ', f_step)
                    move_dict.update({'f_rel':f_step})

                ''' The pauseflag allows:
                    - pausing running acquisitions
                    - wait for slow hardware to catch up (e.g. slow stages)
                '''
                self.move_relative(move_dict)

                while self.pauseflag is True:
                    time.sleep(0.02)
                    QtWidgets.QApplication.processEvents()
                
                QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 1)
                self.image_count += 1

                ''' Keep track of passed time and predict remaining time '''
                time_passed = time.time() - self.start_time
                time_remaining = self.state['predicted_acq_list_time'] - time_passed

                ''' If the time to set up everything is longer than the predicted 
                acq time, the remaining time turns negative - here a different 
                calcuation should be employed here: '''
                if time_remaining < 0:
                    time_passed = time.time() - self.image_acq_start_time
                    time_remaining = self.state['predicted_acq_list_time'] - time_passed

                self.state['remaining_acq_list_time'] = time_remaining
                framerate = self.image_count / time_passed

                ''' Every 100 images, update the predicted acquisition time '''
                if self.image_count % 100 == 0:
                    framerate = self.image_count / time_passed
                    self.state['predicted_acq_list_time'] = self.total_image_count / framerate

                self.send_progress(self.acquisition_count,
                                   self.total_acquisition_count,
                                   i,
                                   steps,
                                   self.total_image_count,
                                   self.image_count,
                                   convert_seconds_to_string(time_passed),
                                   convert_seconds_to_string(time_remaining))
        self.laserenabler.disable_all()
        self.image_acq_end_time = time.time()
        self.image_acq_end_time_string = time.strftime("%Y%m%d-%H%M%S")

        self.close_shutters()

    def close_acquisition(self, acq, acq_list):
        self.sig_status_message.emit('Closing Acquisition: Saving data & freeing up memory')
        if self.stopflag is False:
            # self.move_absolute(acq.get_startpoint(), wait_until_done=True)
            self.close_image_series()
            self.sig_end_image_series.emit(acq, acq_list)

        if self.TTL_mode_enabled_in_cfg is True:
            time.sleep(0.05) # add some buffer time for serial execution
            self.sig_state_request.emit({'ttl_movement_enabled_during_acq' : False})
            time.sleep(0.05)  # buffer time

        # resume asking stages about their position
        self.sig_polling_stage_position_start.emit()

        self.acq_end_time = time.time()
        img_total_time = self.image_acq_end_time - self.image_acq_start_time
        self.acq_end_time_string = time.strftime("%Y%m%d-%H%M%S")
        self.acquisition_count += 1

    @QtCore.pyqtSlot(str)
    def execute_script(self, script):
        self.sig_update_gui_from_state.emit(True)
        self.state['state']='running_script'
        try:
            exec(script)
        except:
            traceback.print_exc()
        self.sig_finished.emit()
        self.state['state']='idle'
        self.sig_update_gui_from_state.emit(False)

    def lightsheet_alignment_mode(self):
        '''Switches shutters after each image to allow coalignment of both lightsheets'''
        self.stopflag = False
        self.sig_prepare_live.emit()
        '''Needs more careful adjustment of the timing

        TODO: There is no wait period to wait for the shutters to open. Nonetheless, the
        visual of the mode impression is not too bad.
        '''
        while self.stopflag is False:
            self.shutter_left.open()
            self.snap_image()
            self.sig_get_live_image.emit()
            self.shutter_left.close()
            ''' Slow down switching to account for slow flip mirror '''
            self.shutter_right.open()
            if self.shutterswitch is True:
                time.sleep(0.25)
                self.shutter_left.open()
            self.snap_image()
            self.sig_get_live_image.emit()
            if self.shutterswitch is True:
                self.shutter_left.close()
            self.shutter_right.close()
            if self.shutterswitch is True:
                time.sleep(0.25)

            QtWidgets.QApplication.processEvents()

        self.close_shutters()
        self.sig_end_live.emit()
        self.sig_finished.emit()

    def visual_mode(self):
        old_l_amp = self.state['etl_l_amplitude']
        old_r_amp = self.state['etl_r_amplitude']
        self.sig_state_request.emit({'etl_l_amplitude' : 0})
        self.sig_state_request.emit({'etl_r_amplitude' : 0})
        time.sleep(0.05)

        self.sig_prepare_live.emit()

        self.stopflag = False

        self.open_shutters()
        while self.stopflag is False:
            ''' Needs update to use snap image in series '''
            self.snap_image()
            self.sig_get_live_image.emit()
            QtWidgets.QApplication.processEvents()

            ''' How to handle a possible shutter switch?'''
            self.open_shutters()

        self.close_shutters()
        self.sig_end_live.emit()

        self.sig_finished.emit()

        self.sig_state_request.emit({'etl_l_amplitude' : old_l_amp})
        self.sig_state_request.emit({'etl_r_amplitude' : old_r_amp})

    def execute_galil_program(self):
        '''Little helper method to execute the program loaded onto the Galil stage:
        allows hand controller to operate'''
        self.sig_state_request.emit({'stage_program' : 'execute'})

    @QtCore.pyqtSlot(str)
    def send_status_message_to_gui(self, string):
        self.sig_status_message.emit(string)

    def list_to_string_with_carriage_return(self, input_list):
        mystring = ''
        for i in input_list:
            mystring = mystring + ' \n ' + i    
        return mystring

    def read_config_parameter(self, key, dictionary):
        """Helper method to check if key exists in the dictionary and read its value, or throw a meaningful error"""
        if key not in dictionary.keys():
            message = f"Mandatory parameter {key} not found in dictionary: \n {dictionary} \n" \
                      f"Check config file for missing parameter {key}."
            logger.error(message)
            raise ValueError(message)
        else:
            return dictionary[key]



'''
acquisitions.py
========================================

Helper classes for mesoSPIM acquisitions
'''

import indexed
import os.path

class Acquisition(indexed.IndexedOrderedDict):
    '''
    Custom acquisition dictionary. Contains all the information to run a single
    acquisition.

    Args:
        x_pos (float): X start position in microns
        y_pos (float): Y start position in microns
        z_start (float): Z start position in microns
        z_end (float): Z end position in microns
        z_step (float): Z stepsize in microns ,
        theta_pos (float): Rotation angle in microns
        f_pos (float): Focus position in microns
        laser (str): Laser designation
        intensity (int): Laser intensity in 0-100
        filter (str): Filter designation (has to be in the config)
        zoom (str): Zoom designation
        filename (str): Filename for the file to be saved

    Attributes:

    Note:
        Getting keys: ``keys = [key for key in acq1.keys()]``

    Example:
        Getting keys: ``keys = [key for key in acq1.keys()]``

    Todo:
        Testtodo-Entry

    '''

    def __init__(self,
                 x_pos=0,
                 y_pos=0,
                 z_start=0,
                 z_end=100,
                 z_step=10,
                 planes=10,
                 theta_pos=0,
                 f_start=0,
                 f_end=0,
                 laser='488 nm',
                 intensity=0,
                 filter='Empty-Alignment',
                 zoom='1x',
                 shutterconfig='Left',
                 folder='tmp',
                 filename='one.tif',
                 etl_l_offset=0,
                 etl_l_amplitude=0,
                 etl_r_offset=0,
                 etl_r_amplitude=0,
                 processing=''):

        super().__init__()

        self['x_pos']=x_pos
        self['y_pos']=y_pos
        self['z_start']=z_start
        self['z_end']=z_end
        self['z_step']=z_step
        self['planes']=planes
        self['rot']=theta_pos
        self['f_start']=f_start
        self['f_end']=f_end
        self['laser']=laser
        self['intensity']=intensity
        self['filter']=filter
        self['zoom']=zoom
        self['shutterconfig']=shutterconfig
        self['folder']=folder
        self['filename']=filename
        self['etl_l_offset']=etl_l_offset
        self['etl_l_amplitude']=etl_l_amplitude
        self['etl_r_offset']=etl_r_offset
        self['etl_r_amplitude']=etl_r_amplitude
        self['processing']=processing


    def __setitem__(self, key, value):
        super().__setitem__(key, value)

    def __call__(self, index):
        ''' This way the dictionary is callable with an index '''
        return self.values()[index]

    def get_keylist(self):
        ''' A list keys is returned for usage as a table header '''
        return [key for key in self.keys()]

    def get_capitalized_keylist(self):
        ''' Here, a list of capitalized keys is returned for usage as a table header '''
        return [key.capitalize() for key in self.keys()]

    def get_image_count(self):
        '''
        Method to return the number of planes in the acquisition
        '''
        return abs(int((self['z_end'] - self['z_start'])/self['z_step']))

    def get_acquisition_time(self, framerate):
        '''
        Method to return the time the acquisition will take at a certain 
        framerate.

        Args:
            float: framerate Framerate of the microscope 

        Returns:
            float: Acquisition time in seconds
        '''

        return self.get_image_count()/framerate

    def get_delta_z_and_delta_f_dict(self, inverted = False):
        ''' Returns relative movement dict for z- and f-steps '''
        if self['z_end'] > self['z_start']:
            z_rel = abs(self['z_step'])
        else:
            z_rel = -abs(self['z_step'])

        ''' Calculate f-step '''
        image_count = self.get_image_count()
        f_rel = abs((self['f_end'] - self['f_start'])/image_count)
        if self['f_end'] < self['f_start']:
            f_rel = -f_rel
        
        if not inverted:
            return {'x_rel' : 0, 'y_rel': 0, 'z_rel' : z_rel, 'f_rel' : f_rel, 'theta_rel': 0}
        else:
            return {'x_rel' : 0, 'y_rel': 0, 'z_rel' : -z_rel, 'f_rel' : -f_rel, 'theta_rel': 0}

    def get_delta_dict(self):
        ''' Returns relative movement dict for z-steps and f-steps'''

        ''' Calculate z-step '''
        if self['z_end'] > self['z_start']:
            z_rel = abs(self['z_step'])
        else:
            z_rel = -abs(self['z_step'])

        ''' Calculate f-step
        image_count = self.get_image_count()
        f_rel = abs((self['f_end'] - self['f_start'])/image_count)
        if self['f_end'] < self['f_start']:
            f_rel = -f_rel
        '''

        return {'z_rel' : z_rel}

    def get_startpoint(self):
        '''
        Provides a dictionary with the startpoint coordinates
        '''
        return {'x_abs': self['x_pos'],
                'y_abs': self['y_pos'],
                'z_abs': self['z_start'],
                'theta_abs': self['rot'],
                'f_abs': self['f_start'],
                }

    def get_endpoint(self):
        return {'x_abs': self['x_pos'],
                'y_abs': self['y_pos'],
                'z_abs': self['z_end'],
                'theta_abs': self['rot'],
                'f_abs': self['f_end'],
                }

    def get_focus_stepsize_generator(self):
        ''''
        Provides a generator object to correct rounding errors for focus tracking acquisitions.

        The focus stage has to travel a shorter distance than the sample z-stage, ideally only
        a fraction of the z-step size. However, due to the limited minimum step size of the focus stage,
        rounding errors can accumulate over thousands of steps.

        Therefore, the generator tracks the rounding error and applies correcting steps here and there
        to minimize the error.

        This assumes a minimum step size of around 0.1 micron that the focus stage is capable of.

        This method contains lots of round functions to keep residual rounding errors at bay.
        '''
        steps = self.get_image_count()
        f_step = abs((self['f_end'] - self['f_start'])/steps)
        if self['f_end'] < self['f_start']:
            f_step = -f_step

        standard_f_step = round(f_step , 1) # Minimum step size: 0.1 micron
        focus_error = 0
        expected_focus = 0
        focus = 0
        for i in range(steps):
            if abs(focus_error) < 0.1:
                focus += standard_f_step
                yield standard_f_step
            else:
                if focus_error < 0:
                    focus += standard_f_step - 0.1
                    yield round(standard_f_step - 0.1,1)
                else:
                    focus += standard_f_step + 0.1
                    yield round(standard_f_step + 0.1,1)
            focus = round(focus, 5)
            expected_focus += f_step
            focus_error = round(expected_focus - focus, 5)


class AcquisitionList(list):
    '''
    Class for a list of acquisition objects

    Examples: "([acq1,acq2,acq3])" is due to the fact that list takes only a single argument
    acq_list = AcquisitionList([acq1,acq2,acq3])
    acq_list.time()
    > 3600
    acq_list.planes()
    > 18000

    acq_list[2](2)
    >10
    acq_list[2]['y_pos']
    >10
    acq_list[2]['y_pos'] = 34


    '''
    def __init__(self, *args):
        list.__init__(self, *args)

        ''' If no arguments are provided, create a
        default acquistion in the list '''

        if len(args) == 0:
            ''' Use a default acquistion '''
            self.append(Acquisition())

        # '''
        # In addition to the list of acquisition objects, the AcquisitionList also
        # contains a rotation point that is save to rotate the sample to the target
        # value.
        # '''
        # self.rotation_point = {'x_abs' : None, 'y_abs' : None, 'z_abs' : None}

    def get_capitalized_keylist(self):
        return self[0].get_capitalized_keylist()

    def get_keylist(self):
        '''
        Here, a list of capitalized keys is returned for usage as a table header
        '''
        return self[0].get_keylist()

    def get_acquisition_time(self, framerate):
        '''
        Returns total time in seconds of a list of acquisitions
        '''
        time = 0
        for i in range(len(self)):
            time += self[i].get_acquisition_time(framerate)

        return time

    def get_image_count(self):
        '''
        Returns the total number of planes for a list of acquistions
        '''
        image_count = 0
        for i in range(len(self)):
            image_count += self[i].get_image_count()

        return image_count

    def get_startpoint(self):
        return self[0].get_startpoint()

    # def set_rotation_point(self, dict):
    #     self.rotation_point = {'x_abs' : dict['x_abs'], 'y_abs' : dict['y_abs'], 'z_abs':dict['z_abs']}

    # def delete_rotation_point(self):
    #     self.rotation_point = {'x_abs' : None, 'y_abs' : None, 'z_abs' : None}

    # def get_rotation_point_status(self):
    #     ''' Returns True if an rotation point was set, otherwise False '''
    #     if self.rotation_point['x_abs'] == None :
    #         return False
    #     else:
    #         return True

    # def get_rotation_point(self):
    #     return self.rotation_point

    def has_rotation(self):
        '''
        Returns true if there is a single rotation in the acq_list.

        TODO: Better method name
        '''
        delta_rot = 0
        for i in range(len(self)-1):
            ''' self[i] is a acq_list element - an acquisition
                get_startpoint() returns the startpoint dict
                dict['theta_abs'] returns the start angle
            '''
            delta_rot = self[i+1].get_startpoint()['theta_abs']-self[i].get_startpoint()['theta_abs']
            if delta_rot != 0:
                return True
        return False

    def get_all_filenames(self):
        ''' Returns a list of all filenames '''
        filename_list = []
        for i in range(len(self)):
            filename = self[i]['folder']+'/'+self[i]['filename']
            filename_list.append(filename)
        return filename_list

    def check_for_existing_filenames(self):
        ''' Returns a list of existing filenames '''
        filename_list = []
        for i in range(len(self)):
            filename = self[i]['folder']+'/'+self[i]['filename']
            file_exists = os.path.isfile(filename)
            if file_exists:
                filename_list.append(filename)
        return filename_list

    def check_filename_extensions(self):
        '''Returns files that have no extension, so their format is undefined.'''
        filename_list = []
        for i in range(len(self)):
            filename = self[i]['filename']
            ext = os.path.splitext(filename)[1]
            if ext == '':
                filename_list.append(filename)
        return filename_list

    def check_for_duplicated_filenames(self):
        ''' Returns a list of duplicated filenames '''
        filenames = []
        # Create a list of full file paths
        for i in range(len(self)):
            if self[i]['filename'][-3:] != '.h5':
                filename = self[i]['folder']+'/'+self[i]['filename']
                filenames.append(filename)
        duplicates = self.get_duplicates_in_list(filenames)

        return duplicates

    def check_for_nonexisting_folders(self):
        ''' Returns a list of nonexisting folders '''
        nonexisting_folders = []
        for i in range(len(self)):
            folder = self[i]['folder']
            if not os.path.isdir(folder):
                nonexisting_folders.append(folder)
        
        return nonexisting_folders

    def get_duplicates_in_list(self, in_list):
        duplicates = []
        unique = set(in_list)
        for each in unique:
            count = in_list.count(each)
            if count > 1:
                duplicates.append(each)
        return duplicates

    def get_n_shutter_configs(self):
        """Get the number of unique shutter configs (1 or 2)"""
        sconfig_list = [a['shutterconfig'] for a in self]
        sconfig_set = set(sconfig_list)
        return len(sconfig_set)

    def get_n_angles(self):
        """Get the number of unique angles"""
        angle_list = [a['rot'] for a in self]
        angle_set = set(angle_list)
        return len(angle_set)

    def get_n_lasers(self):
        """Get the number of unique laser lines"""
        laser_list = [a['laser'] for a in self]
        laser_set = set(laser_list)
        return len(laser_set)

    def get_n_tiles(self):
        """Get the number of tiles as unique (x,y,z_start,rot) combinations"""
        tile_list = []
        for a in self:
            tile_str = f"{a['x_pos']}{a['y_pos']}{a['z_start']}{a['rot']}"
            if not tile_str in tile_list:
                tile_list.append(tile_str)
        return len(tile_list)

    def get_tile_index(self, acq):
        """Get the the tile index for given acquisition"""
        acq_str = f"{acq['x_pos']}{acq['y_pos']}{acq['z_start']}{acq['rot']}"
        tile_list = []
        for a in self:
            tile_str = f"{a['x_pos']}{a['y_pos']}{a['z_start']}{a['rot']}"
            if not tile_str in tile_list:
                tile_list.append(tile_str)
        return tile_list.index(acq_str)

    def get_unique_attr_list(self, key: str = 'laser') -> list:
        """Return ordered list of acquisition attributes.

        Parameters:
        -----------
            key: str
                One of ('laser', 'shutterconfig', 'rot')

        Returns:
        --------
            List of strings, e.g. ('488', '561') for key='laser', in the order of acquisition.
        """
        attributes = ('laser', 'shutterconfig', 'rot')
        assert key in attributes, f'Key {key} must be one of {attributes}.'
        unique_list = []
        for acq in self:
            if acq[key] not in unique_list:
                unique_list.append(acq[key])
        return unique_list

    def find_value_index(self, value: str = '488 nm', key: str = 'laser'):
        """Find the attribute index in the acquisition list.
        Example:
        al = AcquisitionList([Acquisition(), Acquisition(), Acquisition(), Acquisition()])
        al[0]['laser'] = '488 nm' #
        al[1]['laser'] = '488 nm' # gets removed because non-unique
        al[2]['laser'] = '561 nm' #
        al[3]['laser'] = '637 nm' #
        Output:
        al.find_value_index('488 nm', 'laser') # -> 0
        al.find_value_index('561 nm', 'laser') # -> 1
        al.find_value_index('637 nm', 'laser') # -> 2
        """
        unique_list = self.get_unique_attr_list(key)
        assert value in unique_list, f"Value({value}) not found in list {unique_list}"
        return unique_list.index(value)



